# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import re
import warnings
import multiprocessing as mp
from pathlib import Path
from datetime import datetime as dt

import dask
import xarray as xr
import numpy as np

from tqdm import tqdm
from shapely.geometry import Point

from ascat.read_native.xarray_io import cell_io_catalog
from ascat.read_native.xarray_io import swath_io_catalog

from ascat.read_native.xarray_io import trim_dates
from ascat.read_native.xarray_io import append_to_netcdf

from ascat.utils import Spacecraft

process_warnings = True

class CellFileCollectionStack():
    """Collection of grid cell file collections."""

    def __init__(
            self,
            collections,
            ioclass,
            dupe_window=None,
            dask_scheduler="threads",
            **kwargs
    ):
        """Initialize.

        Parameters
        ----------
        collections: list of str or CellFileCollection
            A path to a cell file collection or a list of paths to cell file
            collections, or a list of CellFileCollection.
        ioclass: str, optional
            Name of the ioclass to use for reading the data. Either this or product_id
            must be specified.
        dupe_window : numpy.timedelta64
            Time difference between two observations at the same location_id below which
            the second observation will be considered a duplicate. Will be set to
            `np.timedelta64("10", "m")` if `None`. Default: `None`
        dask_scheduler : str, optional
            Dask scheduler to use for parallel processing. Default: "threads"
        **kwargs
            Keyword arguments to pass to the `ioclass` initialization.
        """
        self.ioclass = ioclass
        if isinstance(collections, (str, Path)):
            collections = [collections]

        if isinstance(collections[0], (str, Path)):
            all_subdirs = self._get_subdirs(collections)
            self.collections = [
                CellFileCollection(
                    subdir, ioclass=self.ioclass, ioclass_kws=kwargs
                )
                for subdir in all_subdirs
            ]
        else:
            self.collections = collections

        self.dupe_window = dupe_window or np.timedelta64(10, "m")
        self.grids = []
        common_cell_size = None
        self._different_cell_sizes = False

        for coll in self.collections:
            if coll.grid not in self.grids:
                self.grids.append(coll.grid)
            if common_cell_size is None:
                common_cell_size = coll.grid_cell_size
            elif coll.grid_cell_size != common_cell_size:
                self._different_cell_sizes = True

        if dask_scheduler is not None:
            dask.config.set(scheduler=dask_scheduler, memory_limit="20GB")

    @staticmethod
    def _get_subdirs(collections):
        """Return a list of all TERMINAL subdirectories of the given collections.

        Parameters
        ----------
        collections : list of str or Path
            List of paths to collections.
        """
        return [Path(r) for c in collections for (r, d, f) in os.walk(c) if not d]

    @classmethod
    def from_product_id(
            cls,
            collections,
            product_id,
            dupe_window=None,
            dask_scheduler=None
    ):
        """Create a CellFileCollectionStack based on a product_id.

        Returns a CellFileCollectionStack object initialized with an io_class specified
        by `product_id` (case-insensitive).

        Parameters
        ----------
        collections : list of str or CellFileCollection
            A path to a cell file collection or a list of paths to cell file collections,
            or a list of CellFileCollection.
        product_id : str
            ASCAT ID of the cell file collections. Either this or ioclass must be
            specified.
        dupe_window : numpy.timedelta64
            Time difference between two observations at the same location_id below which
            the second observation will be considered a duplicate. Will be set to
            `np.timedelta64("10", "m")` if `None`. Default: `None`
        dask_scheduler : str, optional
            Dask scheduler to use for parallel processing. Will be set to "threads" when
            class is initialized if None. Default: None
        """
        product_id = product_id.upper()
        if product_id in cell_io_catalog:
            io_class = cell_io_catalog[product_id]
        else:
            error_str = f"Product {product_id} not recognized. Valid products are"
            error_str += f" {', '.join(cell_io_catalog.keys())}."
            raise ValueError(error_str)

        return cls(
            collections,
            io_class,
            dupe_window=dupe_window,
            dask_scheduler=dask_scheduler
        )

    def add_collection(self, collections, product_id=None):
        """Add a cell file collection to the stack, based on file path.

        Parameters
        ----------
        collections : str or list of str or CellFileCollection
            Path to the cell file collection to add, or a list of paths.
        product_id : str, optional
            ASCAT ID of the collections to add. Needed if collections is a string or
            list of strings.

        Raises
        ------
        ValueError
            If collections is a string or list of strings and product_id is not given.
        """
        new_idx = len(self.collections)
        if isinstance(collections, (str, Path)):
            collections = [collections]
        if isinstance(collections[0], (str, Path)):
            self.collections.extend(
                CellFileCollection.from_product_id(c, product_id)
                for c in self._get_subdirs(collections)
            )
        elif isinstance(collections[0], CellFileCollection):
            self.collections.extend(collections)
        else:
            raise ValueError(
                "collections must be a list of strings or CellFileCollection objects"
            )

        common_cell_size = None
        for coll in self.collections[new_idx:]:
            if coll.grid not in self.grids:
                self.grids.append(coll.grid)
            if common_cell_size is None:
                common_cell_size = coll.grid_cell_size
            elif coll.grid_cell_size != common_cell_size:
                self._different_cell_sizes = True

    def read(
            self,
            cell=None,
            location_id=None,
            bbox=None,
            geom=None,
            mask_and_scale=True,
            date_range=None,
            **kwargs
    ):
        """Read data for a cell or location_id.

        Parameters
        ----------
        cell : int
            Cell number to read data for.
        location_id : int
            Location ID to read data for.
        bbox : tuple
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates to read data within.
        mask_and_scale : bool, optional
            If True, mask and scale the data according to its `scale_factor` and
            `_FillValue`/`missing_value` before returning. Default: True.
        date_range : tuple of numpy.datetime64, optional
            Start and end dates to read data for.
        **kwargs : dict
            Keyword arguments to pass to the read function of the collection

        Returns
        -------
        xarray.Dataset
            Dataset containing the combined data for the given cell or location_id from
            all the collections in the stack.

        Raises
        ------
        ValueError
            If neither cell nor location_id is given.
        """
        if cell is not None:
            data = self._read_cells(cell, date_range=date_range, **kwargs)
        elif location_id is not None:
            data = self._read_locations(location_id, date_range=date_range, **kwargs)
        elif bbox is not None:
            data = self._read_bbox(bbox, date_range=date_range, **kwargs)
        elif geom is not None:
            data = self._read_geometry(geom, date_range=date_range, **kwargs)
        else:
            raise ValueError("Need to specify either cell, location_id or bbox")

        if data is None:
            return data

        data = data.sortby(["sat_id", "locationIndex", "time"])

        # Deduplicate data
        dupl = np.insert(
            (abs(data["time"].values[1:] -
                 data["time"].values[:-1]) < self.dupe_window),
            0,
            False,
        )
        data = data.sel(obs=~dupl)

        if mask_and_scale:
            return xr.decode_cf(data, mask_and_scale=True)
        return data

    def merge_and_write(
            self,
            out_dir,
            cells=None,
            date_range = None,
            out_cell_size=None,
            processes=8
    ):
        """Merge the data in all the collections by cell, and write each cell to disk.

        Parameters
        ----------
        out_dir : str or Path
            Path to output directory.
        cells : list of int, optional
            Cells to write. If None, write all cells.
        date_range : tuple of numpy.datetime64, optional
            Start and end dates to read data for before writing.
        out_cell_size : tuple, optional
            Size of the output cells in degrees (assumes they are square).
            If None, and the component collections all have the same cell size,
            use that.
        processes : int, optional
            Number of processes to use for parallel processing. Default: 8

        Raises
        ------
        ValueError
            If out_cell_size is None and the component collections do not all have the
            same cell size.
        """
        if out_cell_size is None and self._different_cell_sizes is True:
            raise ValueError("Different cell sizes found, need to specify out_cell_size"
                             + " as argument to write_cells function")
        out_dir = Path(out_dir)
        out_dir.mkdir(exist_ok=True, parents=True)
        cells = self.subcollection_cells(cells, out_cell_size, date_range=date_range)

        if processes < 2:
            for cell in tqdm(cells):
                self._write_single_cell(out_dir, self.ioclass, cell, date_range, out_cell_size)
            return

        with mp.Pool(processes=processes) as pool:
            chunksize_heuristic = (len(cells)//processes)+1
            args_list = [
                (out_dir, self.ioclass, cell, date_range, out_cell_size) for cell in cells
            ]

            for _ in tqdm(
                pool.imap_unordered(
                    self._write_single_cell_wrapper,
                    args_list,
                    chunksize=chunksize_heuristic
                ),
                total=len(cells)
            ):
                pass

    # def append_to_collection_on_disk(
    #         self,
    #         disk_collection,
    #         cells=None,
    #         date_range = None,
    #         out_cell_size=None,
    #         processes=8
    # ):
    #     """Append data from the stack to a collection that already exists on disk.

    #     If there is an existing collection on disk with a locations dimension that includes
    #     all the locations in the stack (or the specified selection from the stack), then
    #     data can be appended to that collection without actually reading it into memory.

    #     This is especially useful if a very large collection is used as an archive and
    #     needs to be expanded with new data periodically.

    #     Currently the only supported file format is netCDF4.

    #     You would probably be better served by appending directly from swath files...

    #     Notes
    #     -----
    #     Should take a disk_collection as argument, as well as all the possible arguments
    #     to self.read(), except for (?)

    #     Can't check for duplicates with already-existing data, need to get a max date from
    #     the disk_collection and ensure that only new data is appended.

    #     """
    #     disk_path = disk_collection.path
    #     # should parallelize this
    #     for cell in cells:
    #         # IF cell is not in the disk collection, then just write a new file (but warn)
    #         disk_data = disk_collection.read(cell=cell)
    #         disk_location_ids = disk_data["location_id"].values
    #         stack_data = self.read(
    #             cell=cell,
    #             mask_and_scale=False,
    #             date_range=date_range,
    #             # out_cell_size=out_cell_size
    #         )
    #         stack_location_ids = stack_data["location_id"].values
    #         output_location_ids = disk_location_ids
    #         cell_fname = disk_collection.ioclass.fn_format.format(cell)


    #         # if there are any new locations in the stack that aren't on disk...
    #         if ~np.all(stack_location_ids == disk_location_ids):
    #             if ~np.all(np.isin(stack_location_ids, disk_location_ids)):
    #                 # output_location_ids will now be all unique location_ids in the
    #                 # stack and disk collections combined

    #                 stack_locationIndex = stack_data["locationIndex"].values
    #                 output_location_ids = np.unique(np.append(stack_location_ids, disk_location_ids))
    #                 all_stack_location_ids = stack_location_ids[stack_locationIndex]

    #                 # recalculate disk locationIndex
    #                 disk_locationIndex = disk_data["locationIndex"].values
    #                 all_disk_location_ids = disk_location_ids[disk_locationIndex]
    #                 new_disk_locationIndex = np.searchsorted(
    #                     output_location_ids,
    #                     all_disk_location_ids
    #                 )

    #             # recalculate stack locationIndex
    #             new_stack_locationIndex = np.searchsorted(
    #                 output_location_ids,
    #                 all_stack_location_ids
    #             )

    #             # these new locationIndex arrays have been calculated with respect to a
    #             # sorted location_id array, but there's no guarantee that the location_id
    #             # array on disk or in the stack is actually sorted.

    #             # get the locations from the stack that aren't already in the disk collection
    #             new_locations = stack_data.isel(
    #                 locations=(~np.isin(stack_location_ids, disk_location_ids))
    #             )
    #             # append the new locations dim data
    #             append_to_netcdf(
    #                 disk_path / cell_fname,
    #                 new_locations,
    #                 "locations"
    #             )

    #         # next need to sort the locations dimension, rewrite the locationIndex,
    #         # and then append the new obs dim data

    #         else:
    #             disk_collection.close()
    #             append_to_netcdf(
    #                 disk_path / cell_fname,
    #                 stack_data,
    #                 "obs"
    #             )

    def _collections_in_date_range(self, date_range):
        if date_range is None:
            return self.collections

        start_date = np.datetime64(date_range[0], "ns")
        end_date = np.datetime64(date_range[1], "ns")
        return [
            coll
            for coll in self.collections
            if (
                    (coll.min_datetime is None and coll.max_datetime is None)
                    or ((coll.min_datetime < end_date) and (coll.max_datetime > start_date))
            )
        ]


    def subcollection_cells(self, cells=None, out_cell_size=None, date_range=None):
        """Get the cells that are covered by all the subcollections. If out_cell_size is
        passed, then it returns the cells in the new cell-scheme that are covered by the
        subcollections.

        Parameters
        ----------
        cells : list of int, optional
            Cells to check. If None, check all cells.
        out_cell_size : int, optional
            The size of the cells in the new cell-scheme.

        Returns
        -------
        set
            Cells covered by all subcollections.
        """
        if out_cell_size is None:
            covered_cells = {
                c
                for coll in self._collections_in_date_range(date_range)
                for c in coll.cells_in_collection
                if (cells is None or c in cells)
            }
            return covered_cells

        # if we assume the collections all have the same grid etc then we only
        # really need to check the first collection
        # But would be better to redesign so we just pull from a grid or something
        # Or return a list for each collection... but then the reading logic should
        # be different.
        new_cells = set()
        for coll in self._collections_in_date_range(date_range):
            coll.create_cell_lookup(out_cell_size)
            # add each new cell to the set if any of the old cells that overlap it are
            # in the cells list
            new_cells.update(new_cell for new_cell, old_cells in coll.cell_lut.items()
                             if np.any(np.isin(old_cells,
                                               (cells or coll.cells_in_collection))))

        return new_cells

    def _read_cells(
        self,
        cells,
        search_cell_size=None,
        valid_gpis=None,
        date_range=None,
        **kwargs
    ):
        """Read data for a list of cells.
        If search_cell_size is passed, then it interprets "cells" with respect to the
        new cell-scheme, determines the cells from the original cell-scheme that are
        covered by the new cells, reads those, and trims the extraneous grid points from
        each new cell.

        Parameters
        ----------
        cells : int or list of int
            The cell or list of cells to read.
        search_cell_size : numeric
            The side length, in degrees, of the cell-scheme that the input passed to
            `cells` refers to. That is, if the data is in a 10 degree grid, but
            `search_cell_size` is 5, then passing `cells=13` will result in the function
            returning data from cell 13 of the 5-degree grid, NOT the 10-degree grid.
        valid_gpis : list of int, optional
            List of gpis to keep. If None, keep all gpis.
        date_range : tuple of numpy.datetime64, optional
            Start and end dates to read data for.
        **kwargs
            Keyword arguments to pass to a `CellFileCollection`'s `read` method.

        Returns
        -------
        xarray.Dataset
            Dataset containing the data for the given cells in indexed ragged array
            format.
        """
        search_cells = cells if isinstance(cells, list) else [cells]
        search_cells = np.unique(search_cells)

        if search_cell_size is not None:
            data = []
            for coll in self._collections_in_date_range(date_range):
                coll.create_cell_lookup(search_cell_size)
                old_cells = [
                    c
                    for cell in search_cells
                    for c in coll.cell_lut[cell]
                    if c in coll.cells_in_collection
                ]

                data.extend(
                    [
                        self._trim_to_gpis(
                            coll.read(
                                cell=cell,
                                mask_and_scale=False,
                                **kwargs
                            ),
                            coll.grid.grid_points_for_cell(cell)[0].compressed(),
                        )
                        for cell in old_cells
                    ]
                )

        else:
            data = [
                coll.read(
                    cell=cell,
                    mask_and_scale=False,
                    **kwargs
                )
                for coll in self._collections_in_date_range(date_range)
                for cell in search_cells
            ]

        data = [ds for ds in data if ds is not None]

        if data == []:
            return None

        locs_merged = xr.combine_nested(
            [self._only_locations(ds) for ds in data], concat_dim="locations"
        )

        _, idxs = np.unique(locs_merged["location_id"].values, return_index=True)

        location_vars = {
            var: locs_merged[var][idxs]
            for var in locs_merged.variables
        }

        location_sorter = np.argsort(location_vars["location_id"].values)

        locs_merged.close()

        merged_ds = xr.combine_nested(
            [self._preprocess(ds, location_vars, location_sorter)
             for ds in data],
            concat_dim="obs",
            data_vars="minimal",
            coords="minimal",
            combine_attrs="drop_conflicts",
        )

        merged_ds = trim_dates(merged_ds, date_range)
        merged_ds = self._trim_to_gpis(merged_ds, valid_gpis)

        return merged_ds

    def _read_bbox(
            self,
            bbox,
            date_range=None,
            **kwargs
    ):
        location_ids = np.unique(
            [coll.grid.get_bbox_grid_points(*bbox)
             for coll in self._collections_in_date_range(date_range)]
        )
        return self._read_locations(location_ids, date_range=date_range, **kwargs)
        # cells = self.ioclass.grid.gpi2cell(location_ids)
        # return self._read_cells(cells, valid_gpis=location_ids, date_range=date_range, **kwargs)

    def _read_geometry(
            self,
            geom,
            date_range=None,
            **kwargs
    ):
        bbox = geom.bounds
        latmin, latmax, lonmin, lonmax = bbox[1], bbox[3], bbox[0], bbox[2]
        bbox_gpis, bbox_lats, bbox_lons = self.ioclass.grid.get_bbox_grid_points(
            latmin,
            latmax,
            lonmin,
            lonmax,
            both=True
        )

        # now that we have the grid points that are within the bounding box, we can
        # check which ones are actually within the geometry
        if len(bbox_gpis) > 0:
            geom_location_ids = [
                gpi
                for gpi, lat, lon in zip(bbox_gpis, bbox_lats, bbox_lons)
                if geom.contains(Point(lon, lat))
            ]
        else:
            geom_location_ids = []

        return self._read_locations(geom_location_ids, date_range=date_range, **kwargs)

    def _read_locations(
        self,
        location_ids,
        date_range=None,
        **kwargs
    ):
        cells = self.ioclass.grid.gpi2cell(location_ids)
        return self._read_cells(cells, valid_gpis=location_ids, date_range=date_range, **kwargs)

    # def _read_locations(
    #     self,
    #     location_ids,
    #     date_range=None,
    #     **kwargs
    # ):
    #     location_ids = (
    #         location_ids
    #         if isinstance(location_ids, (list, np.ndarray))
    #         else [location_ids]
    #     )

    #     # all data here is converted to the SAME GRID CELL SIZE within coll.read()
    #     # before being merged later
    #     data = [d for d in
    #             (coll.read(
    #                 location_id=location_id,
    #                 mask_and_scale=False,
    #                 date_range=date_range,
    #                 **kwargs)
    #              for coll in self._collections_in_date_range(date_range)
    #              for location_id in location_ids)
    #             if d is not None]

    #     # merge all the locations-dimensional data vars into one dataset,
    #     # in order to determine the grid points and indexing for the locations
    #     # dimension of the merged dataset.

    #     # coords="all" is necessary in case one of the coords has nan values
    #     # (e.g. altitude, in the case of ASCAT H129)
    #     locs_merged = xr.combine_nested(
    #         [self._only_locations(ds) for ds in data], concat_dim="locations",
    #         coords="all",
    #     )

    #     _, idxs = np.unique(locs_merged["location_id"].values, return_index=True)

    #     location_vars = {
    #         var: locs_merged[var][idxs]
    #         for var in locs_merged.variables
    #     }

    #     location_sorter = np.argsort(location_vars["location_id"].values)

    #     locs_merged.close()

    #     # merge the data variables
    #     merged_ds = xr.combine_nested(
    #         [self._preprocess(ds, location_vars, location_sorter) for ds in data],
    #         concat_dim="obs",
    #         data_vars="minimal",
    #         coords="minimal",
    #     )

    #     return merged_ds

    def _write_single_cell(self, out_dir, ioclass, cell, date_range, out_cell_size, **kwargs):
        """Write data for a single cell from the stack to disk.

        Parameters
        ----------
        out_dir : str or Path
            Path to output directory.
        ioclass : class
            IO class to use for the writer.
        cell : tuple
            Cell to write.
        date_range : tuple of numpy.datetime64
            Start and end dates to read data for before writing.
        out_cell_size : tuple
            Size of the output cell.
        **kwargs
            Keyword arguments to pass to the ioclass write function.
        """
        data = self.read(
            cell=cell,
            date_range=date_range,
            search_cell_size=out_cell_size,
            mask_and_scale=False
        )
        fname = ioclass.fn_format.format(cell)
        data.attrs["id"] = fname
        writer = ioclass(data)
        writer.write(out_dir / fname, **kwargs)
        data.close()
        writer.close()

    def _write_single_cell_wrapper(self, args):
        """Wrap arguments for `self._write_single_cell` and pass to that method.

        Parameters
        ----------
        args : tuple
            Tuple of arguments to pass to `self._write_single_cell`.
        """
        self._write_single_cell(*args)

    @staticmethod
    def _preprocess(ds, location_vars, location_sorter):
        """Pre-processing to be done on a component dataset so it can be merged with others.

        Assumes `ds` is an indexed ragged array. (Re)-calculates the `locationIndex`
        values for `ds` with respect to the `location_id` variable for the merged
        dataset, which may include locations not present in `ds`.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.
        location_vars : dict
            Dictionary of ordered location variable DataArrays for the merged data.
        location_sorter : numpy.ndarray
            Result of `np.argsort(location_vars["location_id"])`, used to calculate
            the `locationIndex` variable. Calculated outside this function to avoid
            re-calculating it for every dataset being merged.

        Returns
        -------
        xarray.Dataset
            Dataset with pre-processing applied.
        """
        # First, we need to calculate the locationIndex variable, based
        # on the location_id variable that will go on the final merged dataset.
        # This should have been stored in self.location_vars["location_id"] at some
        # point before reaching this function, along with all the other
        # locations-dimensional variables in the combined dataset.

        # if "locations" is in the dataset dimensions, then we have
        # a multi-location dataset.
        if "locations" in ds.dims:
            ds = ds.dropna(dim="locations", subset=["location_id"])
            locationIndex = location_sorter[np.searchsorted(
                location_vars["location_id"].values,
                ds["location_id"].values[ds["locationIndex"]],
                sorter=location_sorter,
            )]
            ds = ds.drop_dims("locations")

        # if not, we just have a single location, and logic is different
        else:
            locationIndex = location_sorter[np.searchsorted(
                location_vars["location_id"].values,
                np.repeat(ds["location_id"].values, ds["locationIndex"].size),
                sorter=location_sorter,
            )]

        ds["locationIndex"] = ("obs", locationIndex)

        # Next, we put the locations-dimensional variables on the dataset,
        # and set them as coordinates.
        for var, var_data in location_vars.items():
            ds[var] = ("locations", var_data.values)
        ds = ds.set_coords(["lon", "lat", "alt", "time"])

        try:
            # Need to reset the time index if it's already there, but I can't
            # figure out how to test if time is already an index except like this
            ds = ds.reset_index("time")
        except ValueError:
            pass

        return ds

    @staticmethod
    def _only_locations(ds):
        """Return a dataset with only the variables that aren't in the obs-dimension.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.

        Returns
        -------
        xarray.Dataset
            Dataset with only the locations-dimensional variables.
        """
        return ds[[
            var
            for var in ds.variables
            if "obs" not in ds[var].dims
            and var not in ["row_size", "locationIndex"]
        ]]

    @staticmethod
    def _trim_to_gpis(ds, gpis):
        """Trim a dataset to only the gpis in the given list.
        If any gpis are passed which are not in the dataset, they are ignored.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.
        gpis : list or list-like
            List of gpis to keep.

        Returns
        -------
        xarray.Dataset
            Dataset with only the gpis in the list.

        Notes
        -----
        Does NOT trim the locations dimension, only the obs dimension. There may then
        be empty locations in the dataset.
        """
        if ds is None:
            return None
        if gpis is None:
            return ds

        # first trim out any gpis not in the dataset from the gpi list
        gpis = np.intersect1d(gpis, ds["location_id"].values, assume_unique=True)

        # this is a list of the locationIndex values that correspond to the gpis we're keeping
        locations_idx = np.searchsorted(ds["location_id"].values, gpis)
        # this is the indices of the observations that have any of those locationIndex values
        obs_idx = np.in1d(ds["locationIndex"], locations_idx)

        # now we need to figure out what the new locationIndex vector will be once we drop all the other location_ids
        old_locationIndex = ds["locationIndex"].values
        new_locationIndex = np.searchsorted(
            locations_idx,
            old_locationIndex[np.isin(old_locationIndex, locations_idx)]
        )


        # then trim out any gpis in the dataset not in gpis
        ds = ds.isel({"obs": obs_idx, "locations": locations_idx})
        # and add the new locationIndex
        ds["locationIndex"] = ("obs", new_locationIndex)

        return ds

    def close(self):
        """Close all the collections."""
        for collection in self.collections:
            collection.close()

    def __enter__(self):
        """Context manager initialization."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context related to this object."""
        self.close()

class CellFileCollection:
    """Collection of grid cell files.

    Represents a collection of grid cell files that live in the same directory,
    and contains methods to read data from them.
    """

    def __init__(self,
                 path,
                 ioclass,
                 ioclass_kws=None,
                 dir_name_format="{date1}_{date2}",
                 dir_date_format="%Y%m%d%H%M%S",
                 ):
        """Initialize."""
        self.path = Path(path)
        self.ioclass = ioclass
        self.grid = self.ioclass.grid
        self.grid_cell_size = self.ioclass.grid_cell_size

        self.max_cell = self.ioclass.max_cell
        self.min_cell = self.ioclass.min_cell
        self.dir_name_format = dir_name_format
        self.dir_date_format = dir_date_format
        self.min_datetime, self.max_datetime = self.date_range

        self.fn_format = self.ioclass.fn_format
        self.previous_cells = None
        self.fid = None
        self.out_cell_size = None
        self.cell_lut = None

        if ioclass_kws is None:
            self.ioclass_kws = {}
        else:
            self.ioclass_kws = ioclass_kws

    @property
    def date_range(self):
        """Return the start and end date of the collection based on its dir name"""
        try:
            pattern = braces_to_re_groups(self.dir_name_format)
            match = re.match(pattern, self.path.stem)
            start_date = dt.strptime(match.group("date1"), self.dir_date_format)
            end_date = dt.strptime(match.group("date2"), self.dir_date_format)
            return np.datetime64(start_date, "ns"), np.datetime64(end_date, "ns")
        except Exception:
            warnings.warn(
                f"Could not determine date range for collection '{self.path.stem}'"
                " from directory name."
                " Using min/max datetime from files instead."
            )
            return None, None

    @classmethod
    def from_product_id(cls, collections, product_id, ioclass_kws=None):
        """Create a CellFileCollection based on a product_id.

        Returns a CellFileCollection object initialized with an io_class specified
        by `product_id` (case-insensitive).

        Parameters
        ----------
        collections : list of str or Path
            A path to a cell file collection or a list of paths to cell file collections,
            or a list of CellFileCollection.
        product_id : str
            ASCAT ID of the cell file collections.
        ioclass_kws : dict, optional
            Keyword arguments to pass to the ioclass initialization.

        Raises
        ------
        ValueError
            If product_id is not recognized.
        """
        product_id = product_id.upper()
        if product_id in cell_io_catalog:
            io_class = cell_io_catalog[product_id]
        else:
            error_str = f"Product {product_id} not recognized. Valid products are"
            error_str += f" {', '.join(cell_io_catalog.keys())}."
            raise ValueError(error_str)

        return cls(collections, io_class, ioclass_kws=ioclass_kws)

    @property
    def cells_in_collection(self):
        """Return a list of the cells in the collection.

        Returns
        -------
        list of int
            List of cells in the collection.
        """
        return [int(p.stem) for p in self.path.glob("*")]

    def _write_single_cell(self, out_dir, cell, date_range, out_cell_size, ra_type, **kwargs):
        """Write data for a single cell from the stack to disk.

        Parameters
        ----------
        out_dir : str or Path
            Path to output directory.
        cell : tuple
            Cell to write.
        date_range : tuple of numpy.datetime64
            Start and end dates to read data for before writing.
        out_cell_size : tuple
            Size of the output cell.
        **kwargs
            Keyword arguments to pass to the ioclass write function.
        """
        data = self.read(
            cell=cell,
            date_range=date_range,
            # search_cell_size=out_cell_size,
            mask_and_scale=False
        )
        fname = self.ioclass.fn_format.format(cell)
        data.attrs["id"] = fname
        writer = self.ioclass(data)
        writer.write(out_dir / fname, ra_type=ra_type, **kwargs)
        data.close()
        writer.close()

    def _write_single_cell_wrapper(self, args):
        """Wrap arguments for `self._write_single_cell` and pass to that method.

        Parameters
        ----------
        args : tuple
            Tuple of arguments to pass to `self._write_single_cell`.
        """
        self._write_single_cell(*args)

    def to_contiguous(self, out_dir, out_cell_size, processes=8):
        """
        """
        out_dir = Path(out_dir)
        out_dir.mkdir(exist_ok=True, parents=True)

        cells = self.cells_in_collection
        if processes < 2:
            for cell in tqdm(cells):
                self._write_single_cell(out_dir, cell, None, out_cell_size, "contiguous")
            return

        with mp.Pool(processes=processes) as pool:
            chunksize_heuristic = (len(cells)//processes)+1
            args_list = [
                (out_dir, cell, None, out_cell_size, "contiguous") for cell in cells
            ]

            for _ in tqdm(
                pool.imap_unordered(
                    self._write_single_cell_wrapper,
                    args_list,
                    chunksize=chunksize_heuristic
                ),
                total=len(cells)
            ):
                pass


    def read(
            self,
            cell=None,
            location_id=None,
            coords=None,
            bbox=None,
            geom=None,
            mask_and_scale=True,
            date_range=None,
            **kwargs
    ):
        """Read data from the collection for a cell, location_id, or set of coordinates.

        Parameters
        ----------
        cell : int
            Grid cell number to read.
        location_id : int
            Location id.
        coords : tuple
            Tuple of (lat, lon) coordinates.
        bbox : tuple
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates.
        mask_and_scale : bool, optional
            If True, mask and scale the data according to its `scale_factor` and
            `_FillValue`/`missing_value` before returning. Default: True.
        **kwargs : dict
            Keyword arguments passed to the ioclass.

        Returns
        -------
        xarray.Dataset
            Dataset containing the data for the given cell, location_id, or coordinates.

        Raises
        ------
        ValueError
            If neither cell, location_id, nor coords is given.
        """
        kwargs = {**self.ioclass_kws, **kwargs}
        if cell is not None:
            data = self._read_cell(cell, date_range=date_range, **kwargs)
        elif location_id is not None:
            data = self._read_location_id(location_id, date_range=date_range, **kwargs)
        elif coords is not None:
            data = self._read_latlon(coords[0], coords[1], date_range=date_range, **kwargs)
        elif bbox is not None:
            data = self._read_bbox(bbox, date_range=date_range, **kwargs)
        elif geom is not None:
            data = self._read_geometry(geom, date_range=date_range, **kwargs)
        else:
            raise ValueError("Either cell, location_id or coords (lon, lat)"
                             " must be given")

        if mask_and_scale:
            return xr.decode_cf(data, mask_and_scale=True)

        return data

    def create_cell_lookup(self, out_cell_size):
        """Create a lookup table self.cell_lut mapping a new cell-size grid to the existing one.

        Format of the table is a dictionary, where the keys are the cell numbers
        in the new cell-size grid, and the values are the cell numbers in the
        old cell-size grid which the new cell overlaps.

        Parameters
        ----------
        out_cell_size : int
            Cell size of the new grid.
        """
        if out_cell_size != self.out_cell_size:
            self.out_cell_size = out_cell_size
            new_grid = self.grid.to_cell_grid(out_cell_size)
            old_cells = self.grid.arrcell
            new_cells = new_grid.arrcell
            # old_cells and new_cells are arrays of the same length, with each element
            # corresponding to a grid point in self.grid, its value representing the
            # cell that grid point is in.
            self.cell_lut = {new_cell:
                             np.unique(old_cells[np.where(new_cells == new_cell)[0]])
                             for new_cell in np.unique(new_cells)}

    def _open(self, location_id=None, cells=None, grid=None):
        """Open cell file using the given location_id (will open the cell file containing
        the location_id) or cell number. Sets self.fid to an xarray dataset representing
        the open cell file.

        Parameters
        ----------
        location_id : int (optional)
            Location identifier.
        cell : int (optional)
            Cell number.
        grid : pygeogrids.CellGrid (optional)
            Grid object.

        Returns
        -------
        success : boolean
            Flag if opening the file was successful.
        """
        success = True
        if location_id is not None:
            grid = grid or self.grid
            cells = grid.gpi2cell(location_id)

        if not isinstance(cells, (list, np.ndarray)):
            cells = [cells]

        filenames = [self.get_cell_path(c) for c in np.unique(cells)]

        if len(filenames) == 1:
            filenames = filenames[0]

        if (self.previous_cells is None) or set(self.previous_cells) != set(cells):
            self.close()

            try:
                self.fid = self.ioclass(filenames, **self.ioclass_kws)
            except IOError as e:
                success = False
                self.fid = None
                msg = f"I/O error({e.errno}): {e.strerror}, {filenames}"
                warnings.warn(msg, RuntimeWarning, stacklevel=2)
                self.previous_cells = None
            else:
                self.previous_cells = cells

        return success

    def _read_cell(self, cell, date_range=None, **kwargs):
        """Read data from one entire cell.

        Parameters
        ----------
        cell : int
            Cell number.
        **kwargs : dict
            Keyword arguments passed to the ioclass.

        Returns
        -------
        data : xarray.Dataset
            Data for the cell.
        """
        data = None
        if self._open(cells=cell):
            data = self.fid.read(mask_and_scale=False,
                                 date_range=date_range,
                                 **kwargs)

        return data

    def _read_cells(self, cell, date_range=None, **kwargs):
        """Read data from several entire cells.

        Parameters
        ----------
        cell : int
            Cell number.
        **kwargs : dict
            Keyword arguments passed to the ioclass.

        Returns
        -------
        data : xarray.Dataset
            Data for the cell.
        """
        data = None
        if self._open(cells=cell):
            data = self.fid.read(mask_and_scale=False,
                                 date_range=date_range,
                                 **kwargs)

        return data

    def _read_latlon(self, lat, lon, date_range=None, **kwargs):
        """Read data for the nearest grid point to the given coordinate.

        Converts lon/lat pair to a location_id, then reads data for that location_id.

        Parameters
        ----------
        lat : float
            Latitude coordinate.
        lon : float
            Longitude coordinate.
        **kwargs : dict
            Keyword arguments passed to the ioclass `read` method.

        Returns
        -------
        data : xarray.Dataset
            Data at the given coordinates.
        """
        location_id, _ = self.grid.find_nearest_gpi(lon, lat)

        return self._read_location_id(location_id,
                                      date_range=date_range,
                                      **kwargs)

    def _read_bbox(self, bbox, date_range=None, **kwargs):
        location_ids = self.grid.get_bbox_grid_points(*bbox)
        return self._read_location_id(location_ids, date_range=date_range, **kwargs)

    def _read_geometry(self, geom, date_range=None, **kwargs):
        bbox = geom.bounds
        latmin, latmax, lonmin, lonmax = bbox[1], bbox[3], bbox[0], bbox[2]
        bbox_gpis, bbox_lats, bbox_lons = self.grid.get_bbox_grid_points(
            latmin,
            latmax,
            lonmin,
            lonmax,
            both=True
        )

        if len(bbox_gpis) > 0:
            geom_location_ids = [
                gpi
                for gpi, lat, lon in zip(bbox_gpis, bbox_lats, bbox_lons)
                if geom.contains(Point(lon, lat))
            ]
        else:
            geom_location_ids = []

        return self._read_location_id(geom_location_ids, date_range=date_range, **kwargs)

    def _read_location_id(self, location_id, date_range=None, **kwargs):
        """Read data for given grid point.

        Parameters
        ----------
        location_id : int
            Location identifier.
        **kwargs : dict
            Keyword arguments passed to the ioclass `read` method.

        Returns
        -------
        data : xarray.Dataset
            Data for the given grid point.
        """
        data = None

        if self._open(location_id=location_id):
            data = self.fid.read(location_id=location_id,
                                 mask_and_scale=False,
                                 date_range=date_range,
                                 **kwargs)

        return data

    def get_cell_path(self, cell=None, location_id=None):
        """Get path to cell file given cell number or location id.

        Returns a path to a cell file in the collection's directory, whether the file
        exists or not, as long as the cell number or location id is within the grid.

        Parameters
        ----------
        cell : int, optional
            Cell number.
        location_id : int, optional
            Location identifier.

        Returns
        -------
        path : pathlib.Path
            Path to cell file.

        Raises
        ------
        ValueError
            If neither cell nor location_id is given.
        ValueError
            If the given cell number or location_id is not within the grid.
        """
        if location_id is not None:
            cell = self.grid.gpi2cell(location_id)
        elif cell is None:
            raise ValueError("Either location_id or cell must be given")

        if (cell > self.max_cell) or (cell < self.min_cell):
            raise ValueError(f"Cell {cell} is not in grid")

        return self.path / self.fn_format.format(cell)

    def close(self):
        """Close file."""
        if self.fid is not None:
            self.fid.close()
            self.fid = None

    def __enter__(self):
        """Context manager initialization."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context related to this object."""
        self.close()


class SwathFileCollection:
    """ Collection of time-series swath files.

    Parameters
    ----------
    path : str or Path
        Path to the swath file collection.
    ioclass : ascat.read_native.xarray_io.SwathIOBase
        IO class to use for reading the data.
    ioclass_kws : dict, optional
        Keyword arguments to pass to the ioclass initialization. Default: None
    dask_scheduler : str, optional
        Dask scheduler to use for parallel processing in xarray. In testing this just
        made most things slower, but it may be useful in some cases. Default: None


    Attributes
    ----------
    path : Path
        Path to the swath file collection.
    ioclass : class
        IO class to use for reading the data.
    ioclass_kws : dict
        Keyword arguments to pass to the ioclass initialization. May include ioclass
        attributes that will override any that are set in the current ioclass.
    grid : pygeogrids.CellGrid object
        Grid object defining the grid the data is on.
    ts_dtype : numpy.dtype
        Data types to encode the time series data as when writing.
    beams_vars : list of str
        List of names of the variables that have a beams dimension.
    date_format : str
        Format of the date in the filename.
    cell_fn_format : str
        Format for the names of the cell files that will be written out.
    chron_files : function
        Function to search for files in the collection based on their date.
    previous_cell : int or list of int
    fid : ascat.read_native.xarray_io.SwathIOBase object
        The currently open instance of self.ioclass.
    max_buffer_memory_mb : int
        Maximum amount of memory to use for buffering data when stacking to disk.

    Methods
    -------
    """

    def __init__(self,
                 path,
                 ioclass,
                 ioclass_kws=None,
                 dask_scheduler=None,
                 ):
        self.path = Path(path)
        self.ioclass = ioclass

        if ioclass_kws is None:
            self.ioclass_kws = {}
        else:
            self.ioclass_kws = ioclass_kws
            for k in list(ioclass_kws.keys()):
                # allow user to override individual ioclass attributes if they want
                if k in [
                        "fn_pattern",
                        "sf_pattern",
                        "date_format",
                        "grid",
                        "grid_cell_size",
                        "cell_fn_format",
                        "beams_vars",
                        "ts_dtype",
                        "fn_read_fmt",
                        "sf_read_fmt",
                ]:
                    v = ioclass_kws.pop(k)
                    setattr(self.ioclass, k, v)

        self.grid = self.ioclass.grid
        self.ts_dtype = self.ioclass.ts_dtype
        self.beams_vars = self.ioclass.beams_vars

        self.date_format = self.ioclass.date_format
        self.cell_fn_format = self.ioclass.cell_fn_format
        # this should perhaps be optional, the user could provide their own filesearch function
        self.chron_files = self.ioclass.chron_files(self.path)

        self.previous_cell = None
        self._open_fnames = None
        self.fid = None
        self.max_buffer_memory_mb = 6*1024

        self._possible_gpis = None


        if dask_scheduler is not None:
            dask.config.set(scheduler=dask_scheduler)

    @classmethod
    def from_product_id(cls, path, product_id, ioclass_kws=None, dask_scheduler=None):
        """Create a SwathFileCollection based on a product_id.

        Returns a SwathFileCollection object initialized with an io_class specified
        by `product_id` (case-insensitive).

        Parameters
        ----------
        path : str or Path
            Path to the swath file collection.
        product_id : str
            Identifier for the specific ASCAT product the swath files are part of.
        ioclass_kws : dict, optional
            Keyword arguments to pass to the ioclass initialization. Default: None
        dask_scheduler : str, optional
            Dask scheduler to use for parallel processing. Will be set to "threads" when
            class is initialized if None. Default: None

        Raises
        ------
        ValueError
            If product_id is not recognized.

        Examples
        --------
        >>> my_swath_collection = SwathFileCollection.from_product_id(
        ...     "/path/to/swath/files",
        ...     "H129",
        ... )

        """
        product_id = product_id.upper()
        if product_id in swath_io_catalog:
            io_class = swath_io_catalog[product_id]
        else:
            error_str = f"Product {product_id} not recognized. Valid products are"
            error_str += f" {', '.join(swath_io_catalog.keys())}."
            raise ValueError(error_str)

        return cls(
            path,
            io_class,
            ioclass_kws=ioclass_kws,
            dask_scheduler=dask_scheduler
        )

    def stack(
            self,
            out_dir,
            fnames=None,
            date_range=None,
            mode="w",
            processes=1,
            buffer_memory_mb=None,
            dupe_window=None
    ):
        """Stack swath files and split them into cell timeseries files.

        Reads swath files into memory, stacking their datasets in a buffer until the sum
        of their sizes exceeds self.max_buffer_memory_mb. Then, splits the buffer into
        cell timeseries datasets, writes them to disk in parallel, and clears the
        buffer. This process repeats until all files have been processed, with
        subsequent writes appending new data to existing cell files when appropriate.

        Parameters
        ----------
        out_dir : pathlib.Path
            Output directory to write the stacked files to.
        fnames : list of pathlib.Path, optional
            List of swath filenames to stack.
        date_range : tuple of datetime.datetime
            Start and end dates to read data for before writing.
        mode : str, optional
            Write mode. Default is "w", which will clear all files from out_dir before
            processing. Use "a" to append data to existing files (only if those have
            also been produced by this function).
        processes : int, optional
            Number of processes to use for parallel writing. Default is 1.
        buffer_memory_mb : numeric, optional
            Maximum amount of memory to use for the buffer, in megabytes. Will be set to
            `self.max_buffer_memory_mb` if None. Default is None.
        dupe_window : numpy.timedelta64, optional
            Time window within which duplicate observations will be removed. Default is
            `None`.

        Raises
        ------
        ValueError
            If mode is not "w" or "a".
        """
        if mode == "w":
            if process_warnings == True:
                input(f"Calling ragged_array_ts.stack with mode='w' will clear all files"
                    f" from {out_dir}.\nPress enter to continue, or ctrl+c to cancel. ")
            for f in out_dir.glob("*.nc"):
                f.unlink()
        elif mode == "a":
            if process_warnings == True:
                input(f"Calling ragged_array_ts.stack with mode='a' will append data to"
                    f" existing files in `out_dir`:\n{out_dir}\nIf it is important to"
                    " preserve the data in this directory it its current state, please"
                    " save a backup elsewhere before continuing, or choose a different"
                    " 'out_dir' for this function and combine the results afterwards"
                    " using a CellFileCollectionStack.\nPress enter to continue, or"
                    " ctrl+c to cancel. ")
        else:
            raise ValueError(f"Invalid mode {mode} for 'SwathFileCollection.stack'."
                             " Valid modes are 'w' and 'a'.")

        if date_range is not None:
            start_dt, end_dt = date_range
            range_fnames = self.get_filenames(start_dt=start_dt, end_dt=end_dt)
            if fnames is not None:
                fnames = list(set(fnames).intersection(set(range_fnames)))
                warnings.warn("Both `fnames` and `date_range` were passed. `fnames`"
                              " will be filtered to only include files within the"
                              " passed `date_range`.\n")
            else:
                fnames = range_fnames
        elif fnames is None:
            fnames = self.get_filenames()

        self.max_buffer_memory_mb = buffer_memory_mb or self.max_buffer_memory_mb
        buffer = []
        buffer_size = 0
        # process = psutil.Process()
        total_swaths = len(fnames)
        for iteration, f in enumerate(fnames):
            self._open(f)
            ds = self.fid.read(mask_and_scale=False).load()
            # buffer_size  = process.memory_info().rss / 1e6
            print("Filling swaths buffer..."
                  f" {buffer_size:.2f}MB/{self.max_buffer_memory_mb:.2f}MB",
                  end="\r")
            buffer_size += ds.nbytes / 1e6
            if buffer_size > self.max_buffer_memory_mb:
                out_dir.mkdir(parents=True, exist_ok=True)
                print("\nBuffer full. Processing swath files..."
                      "                     ",
                      end="\r")
                combined_ds = self.process(
                    xr.combine_nested(
                        buffer,
                        concat_dim="obs",
                        combine_attrs=self.ioclass.combine_attributes
                    )
                )
                print(f"Processed {iteration}/{total_swaths} swath files."
                      " Dumping to cell files...")
                self._parallel_write_cells(
                    combined_ds,
                    out_dir,
                    processes=processes,
                    dupe_window=dupe_window
                )
                print("Finished dumping buffer to cell files.")
                combined_ds.close()
                buffer = []
                buffer_size = ds.nbytes / 1e6
            buffer.append(ds)

        if len(buffer) > 0:
            out_dir.mkdir(parents=True, exist_ok=True)
            print(f"Processed {total_swaths}/{total_swaths} swath files."
                  " Processing remaining buffer data...",
                  end="\r")
            combined_ds = self.process(
                xr.combine_nested(
                    buffer,
                    concat_dim="obs",
                    combine_attrs=self.ioclass.combine_attributes,
                )
            )
            print(f"Processed {total_swaths}/{total_swaths} swath files."
                  " Dumping to cell files...")
            self._parallel_write_cells(
                combined_ds,
                out_dir,
                processes=processes,
                dupe_window=dupe_window
            )
            total_cell_files = len(list(out_dir.glob("*.nc")))
            print(f"Finished stacking {total_swaths} swath files to {total_cell_files}"
                  " cell files.")
            combined_ds.close()

    def process(self, data):
        """Process a stacked dataset of swath data into a format that is ready to be
        split into cell timeseries datasets, and return the processed dataset.

        Parameters
        ----------
        data : xarray.Dataset
            Stacked dataset to process.
        """
        # if there are kwargs, use them instead of self.ioclass_kws
        beam_idx = {"for": 0, "mid": 1, "aft": 2}

        # if any beam has backscatter data for a record, the record is valid. Drop
        # observations that don't have any backscatter data.
        if "backscatter" in data.variables:
            if data["obs"].size > 0:
                data = data.sel(
                    obs=~np.all(
                        data["backscatter"] == data["backscatter"].attrs["missing_value"],
                        # np.isnan(data["backscatter"]),
                        axis=1
                    )
                )

        # break the beams dimension variables into separate variables for
        # the fore, mid, and aft beams
        if data["obs"].size > 0:
            if "beams" in data.dims:
                for var in self.ts_dtype.names:
                    if var[:-4] in self.beams_vars:
                        ending = var[-3:]
                        data[var] = data.sel(beams=beam_idx[ending])[var[:-4]]
                    # if var in data.variables:
                    #     if data[var].dtype != self.ts_dtype[var]:
                    #         data[var] = data[var].astype(self.ts_dtype[var])
                        # data[var].attrs["dtype"] = self.ts_dtype[var]
                        # data[var].attrs["_FillValue"] = dtype_to_nan[self.ts_dtype[var]]
                        # data[var].attrs["missing_value"] = data[var].attrs["_FillValue"]

                # drop the variables on the beams dimension
                data = data.drop_dims("beams")

        # Find which cell each observation belongs to, and assign it as a coordinate.
        data = data.assign_coords(
            {"cell": ("obs", self.grid.gpi2cell(data["location_id"].values))}
        )
        # Must set an index for the cell coordinate so that we can select by it later.
        data = data.set_xindex("cell")

        return data

    def read(
            self,
            date_range,
            cell=None,
            location_id=None,
            coords=None,
            bbox=None,
            geom=None,
            **kwargs
    ):
        """Takes either 1 or 2 arguments and calls the correct function
        which is either reading the gpi directly or finding
        the nearest gpi from given lat,lon coordinates and then reading it.

        If the time range is large, this can be slow. It may make more sense to
        convert to cell files first and access that data from disk using
        a `CellFileCollection` or `CellFileCollectionStack`.

        Parameters
        ----------
        date_range : tuple of datetime.datetime
            Start and end dates.
        cell : int or list of int, optional
            Grid cell number to read.
        location_id : int, optional
            Location id.
        coords : tuple, optional
            Tuple of (lat, lon) coordinates.
        bbox : tuple, optional
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates.
        geometry : shapely.geometry, optional
            Geometry object; use to read data that intersects the geometry.
        """
        start_dt, end_dt = date_range
        fnames = self.get_filenames(
            start_dt,
            end_dt,
            cell=cell,
            location_id=location_id,
            coords=coords,
            bbox=bbox,
            geom=geom,
        )
        if cell is not None:
            data = self._read_cell(fnames, cell, **kwargs)
        elif location_id is not None:
            data = self._read_location_id(fnames, location_id, **kwargs)
        elif coords is not None:
            data = self._read_latlon(fnames, coords[0], coords[1], **kwargs)
        elif bbox is not None:
            data = self._read_bbox(fnames, bbox, **kwargs)
        elif geom is not None:
            data = self._read_geometry(fnames, geom, **kwargs)
        elif self._open(fnames):
            data = self.fid.read(**kwargs)
        else:
            raise ValueError(f"No swath files found in directory {self.path} for the"
                             f" passed date range: {start_dt} - {end_dt}")

        data = data.set_xindex("time")
        start_dt = np.datetime64(start_dt)
        end_dt = np.datetime64(end_dt)
        data = data.isel(obs=(data.time >= start_dt) & (data.time < end_dt))
        data = data.reset_index("time", drop=False)
        return data

    def close(self):
        """Close collection and constituent xarray datasets."""
        if self.fid is not None:
            self.fid.close()
            self.fid = None

    def get_filenames(
            self,
            start_dt=None,
            end_dt=None,
            cell=None,
            location_id=None,
            coords=None,
            bbox=None,
            geom=None,
    ):
        """Get filenames for the given time range.

        Parameters
        ----------
        start_dt : datetime.datetime
            Start time.
        end_dt : datetime.datetime
            End time.

        Returns
        -------
        fnames : list of pathlib.Path
            List of filenames.

        Raises
        ------
        NotImplementedError
            If the ioclass does not have a file search method named `chron_files`.
        """
        if start_dt is None and end_dt is None:
            fnames = list(self.path.glob("**/*.nc"))
        elif self.chron_files:
            fnames = self.chron_files.search_period(
                start_dt,
                end_dt,
                date_field_fmt=self.date_format,
                end_inclusive=False
            )
        else:
            raise NotImplementedError("File search not implemented for this product."
                                      " Check if fn_pattern and sf_pattern are defined"
                                      f" in ioclass {self.ioclass.__name__}")


        # now check for location_ids
        location_ids = None
        if cell is not None:
            location_ids = self.grid.grid_points_for_cell(cell)
        elif location_id is not None:
            location_ids = [location_id]
        elif coords is not None:
            location_ids = self._location_id_from_coords(*coords)
        elif bbox is not None:
            location_ids = self._location_id_from_bbox(bbox)
        elif geom is not None:
            location_ids = self._location_id_from_geometry(geom)

        if location_ids is not None:
            gpi_lookup = np.zeros(self.grid.gpis.max()+1, dtype=bool)
            gpi_lookup[location_ids] = 1
            fnames = [f
                    for f in fnames
                    if self._open(f) and self.fid.contains_location_ids(lookup_vector=gpi_lookup)]

        return fnames

    def swath_data_generator(
            self,
            start_dt=None,
            end_dt=None,
            cell=None,
            location_id=None,
            coords=None,
            bbox=None,
            geom=None,
    ):
        """Return a generator producing the data for each requested swath file.

        Parameters
        ----------
        start_dt : datetime.datetime
            Start time.
        end_dt : datetime.datetime
            End time.
        cell : int
            Grid cell number to select.
        location_id : int
            Location id.
        coords : tuple
            Tuple of (lat, lon) coordinates.
        bbox : tuple
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates.
        geom : shapely.geometry
            Geometry object; use to select data that intersects the geometry.

        Yields
        ------
        start_timestamp : numpy.datetime64
            Sensing start time of the swath file.
        end_timestamp : numpy.datetime64
            Sensing end time of the swath file.
        sat : str
            Satellite name.
        data : xarray.Dataset
            Dataset for each swath file intersecting the requested extent.
        """
        fnames = self.get_filenames(
            start_dt=start_dt,
            end_dt=end_dt,
            cell=cell,
            location_id=location_id,
            coords=coords,
            bbox=bbox,
            geom=geom,
        )

        for f in fnames:
            self._open(f)
            data = self.fid.read()
            start_timestamp = np.datetime64(data.attrs["sensing_start_time_utc"], "ns")
            end_timestamp = np.datetime64(data.attrs["sensing_end_time_utc"], "ns")
            spacecraft = Spacecraft(data.attrs["spacecraft"])
            sat = spacecraft.satellite

            yield start_timestamp, end_timestamp, sat, data

    def _open(self, fnames):
        """Open swath files

        Parameters
        ----------
        location_id : int
            Location identifier.

        Returns
        -------
        success : boolean
            Flag if opening the file was successful.
        """
        success = True
        if fnames != self._open_fnames:
            self.fid = None

        if self.fid is None:
            try:
                self.fid = self.ioclass(fnames, **self.ioclass_kws)
            except IOError as e:
                success = False
                self.fid = None
                msg = f"I/O error({e.errno}): {e.strerror}"
                warnings.warn(msg, RuntimeWarning, stacklevel=2)
        # else:
        #     # handling for adding new extra data to the open dataset
        #     pass

        return success

    def _cell_data_as_indexed_ra(self, ds, cell, dupe_window=None):
        """Convert swath data for a single cell to indexed format.

        Parameters
        ----------
        ds : xarray.Dataset
            Input dataset.

        Returns
        -------
        ds : xarray.Dataset
            Output dataset.
        """
        dupe_window = dupe_window or np.timedelta64(10, "m")
        # location_id = self.grid
        ds = ds.drop_vars(["latitude", "longitude", "cell"], errors="ignore")
        location_id, lon, lat = self.grid.grid_points_for_cell(cell)
        sorter = np.argsort(location_id)
        locationIndex = sorter[np.searchsorted(location_id,
                                               ds["location_id"].values,
                                               sorter=sorter)]

        ds = ds.assign_coords({"lon": ("locations", lon),
                               "lat": ("locations", lat),
                               "alt": ("locations", np.repeat(np.nan,
                                                              len(location_id)))})

        ds = ds.set_coords(["time"])
        ds["location_id"] = ("locations", location_id)
        ds["location_description"] = ("locations", np.repeat("", len(location_id)))
        ds["locationIndex"] = ("obs", locationIndex)

        # set _FillValue to missing_value (which has been deprecated)
        for var in ds.variables:
            if "missing_value" in ds[var].attrs and "_FillValue" not in ds[var].attrs:
                ds[var].attrs["_FillValue"] = ds[var].attrs.get("missing_value")

        ds = ds.sortby(["locationIndex", "time"])

        # dedupe
        dupl = np.insert(
            (abs(ds["time"].values[1:] -
                 ds["time"].values[:-1]) < dupe_window),
            0,
            False,
        )
        ds = ds.sel(obs=~dupl)

        return ds

    def _write_cell_ds(self, ds, out_dir, cell, dupe_window=None):
        """Write a cell dataset to a file."""
        fname = self.ioclass.cell_fn_format.format(cell)

        out_ds = self._cell_data_as_indexed_ra(ds, cell, dupe_window)
        # out_ds = xr.decode_cf(out_ds, mask_and_scale=False)
        # for var in self.ts_dtype.names:
        #     if var in out_ds.variables:
        #         if out_ds[var].dtype != self.ts_dtype[var]:
        #             out_ds[var] = out_ds[var].astype(self.ts_dtype[var])

        out_ds = xr.decode_cf(out_ds, mask_and_scale=True)
        out_ds.attrs["id"] = fname

        writer = self.ioclass(out_ds)
        writer.write(out_dir / fname, mode="a")
        writer.close()

    def _write_cell_ds_wrapper(self, args):
        """Wrapper for write_cell_ds to allow parallel processing with imap."""
        self._write_cell_ds(*args)

    def _parallel_write_cells(self, ds, out_dir, processes=8, dupe_window=None):
        """Write a stacked dataset to a set of cell files in parallel."""
        cells = np.unique(ds.cell.values)
        args = [
            (
                ds.isel(obs=np.where(ds.cell.values == cell)[0]),
                out_dir,
                cell,
                dupe_window
            )
            for cell in cells
        ]

        if processes > 1:
            chunksize_heuristic = (len(cells)//processes)+1
            with mp.Pool(processes=processes) as pool:
                for _ in tqdm(pool.imap_unordered(self._write_cell_ds_wrapper,
                                                  args,
                                                  chunksize=chunksize_heuristic),
                              total=len(cells)):
                    pass
        else:
            for arguments in tqdm(args):
                self._write_cell_ds(*arguments)

    def _location_id_from_coords(self, lat, lon):
        """Get the location_id for a given lat/lon pair."""
        return self.grid.find_nearest_gpi(lon, lat)[0]

    def _location_id_from_bbox(self, bbox):
        """Get the location_ids for a given bounding box.
        (latmin, latmax, lonmin, lonmax)"""
        return self.grid.get_bbox_grid_points(*bbox)


    def _location_id_from_geometry(self, geom):
        """Get the location_ids for a given geometry."""
        bbox = geom.bounds
        latmin, latmax, lonmin, lonmax = bbox[1], bbox[3], bbox[0], bbox[2]
        bbox_gpis, bbox_lats, bbox_lons = self.grid.get_bbox_grid_points(
            latmin,
            latmax,
            lonmin,
            lonmax,
            both=True
        )
        # now that we have the grid points that are within the bounding box, we can
        # check which ones are actually within the geometry
        if len(bbox_gpis) > 0:
            geom_location_ids = [
                gpi
                for gpi, lat, lon in zip(bbox_gpis, bbox_lats, bbox_lons)
                if geom.contains(Point(lon, lat))
            ]
        else:
            geom_location_ids = []

        return geom_location_ids

    def _read_cell(self, fnames, cell, **kwargs):
        """Read data from the entire cell."""
        # if there are kwargs, use them instead of self.ioclass_kws

        data = None
        if self._open(fnames):
            data = self.fid.read(cell=cell, **kwargs)

        return data

    def _read_latlon(self, fnames, lat, lon, **kwargs):
        """Reading data for given longitude and latitude coordinate.

        Parameters
        ----------
        lat : float
            Latitude coordinate.
        lon : float
            Longitude coordinate.

        Returns
        -------
        data : dict of values
            data record.
        """
        location_id = self._location_id_from_coords(lat, lon)

        return self._read_location_id(fnames, location_id, **kwargs)

    def _read_bbox(self, fnames, bbox, **kwargs):
        """Reading data for given bounding box.

        Parameters
        ----------
        bbox : tuple or list
            Bounding box coordinates (latmin, latmax, lonmin, lonmax).
        """
        location_ids = self._location_id_from_bbox(bbox)
        return self._read_location_id(fnames, location_ids, **kwargs)

    def _read_geometry(self, fnames, geom, **kwargs):
        """Reading data for given geometry.

        Assumes geometry and grid are in the same coordinate system.

        Parameters
        ----------
        geometry : shapely.geometry
            Geometry object.
        """
        location_ids = self._location_id_from_geometry(geom)
        return self._read_location_id(fnames, location_ids, **kwargs)

    def _read_location_id(self, fnames, location_id, **kwargs):
        """Read data for given grid point.

        Parameters
        ----------
        location_id : int
            Location identifier.

        Returns
        -------
        data : numpy.ndarray
            Data.
        """
        data = None

        if self._open(fnames):
            gpi_lookup = np.zeros(self.grid.gpis.max()+1, dtype=bool)
            gpi_lookup[location_id] = 1
            data = self.fid.read(location_id=location_id, lookup_vector=gpi_lookup, **kwargs)

        return data


    def __enter__(self):
        """Context manager initialization."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context related to this object."""
        self.close()

def braces_to_re_groups(string):
    """
    Convert braces to character patterns defining regular expression groups.
    If any group name is repeated in the template string, a backreference
    is used for subsequent appearances.

    Parameters
    ----------
    string : str
        String with braces.

    Returns
    -------
    string : str
        String with regular expression groups.

    Examples
    --------
    >>> braces_to_re_groups("{year}-{month}-{day}")
    "(?P<year>.+)-(?P<month>.+)-(?P<day>.+)"
    >>> braces_to_re_groups("{year}-{month}-{day}_{year}-{month}-{day2}")
    "(?P<year>.+)-(?P<month>.+)-(?P<day>.+)_(?P=year)-(?P=month)-(?P<day2>.+)"
    """

    pattern = re.compile(r"{(.+?)}")
    seen = set()
    parts = pattern.split(string)

    for i in range(1, len(parts), 2):
        content = parts[i]

        if content in seen:
            parts[i] = f"(?P={content})"
        else:
            parts[i] = f"(?P<{content}>.+)"
            seen.add(content)

    return "".join(parts)

####################################################
class RAFile:
    """
    Base class used for Ragged Array (RA) time series data.
    """

    def __init__(
        self,
        loc_dim_name="locations",
        obs_dim_name="time",
        loc_ids_name="location_id",
        loc_descr_name="location_description",
        time_units="days since 1900-01-01 00:00:00",
        time_var="time",
        lat_var="lat",
        lon_var="lon",
        alt_var="alt",
        cache=False,
        mask_and_scale=False,
    ):
        """
        Initialize.

        Parameters
        ----------
        loc_dim_name : str, optional
            Location dimension name (default: "locations").
        obs_dim_name : str, optional
            Observation dimension name (default: "time").
        loc_ids_name : str, optional
            Location IDs name (default: "location_id").
        loc_descr_name : str, optional
            Location description name (default: "location_description").
        time_units : str, optional
            Time units definition (default: "days since 1900-01-01 00:00:00").
        time_var : str, optional
            Time variable name (default: "time").
        lat_var : str, optional
            Latitude variable name (default: "lat").
        lon_var : str, optional
            Latitude variable name (default: "lon").
        alt_var : str, optional
            Altitude variable name (default: "alt").
        cache : boolean, optional
            Cache flag (default: False).
        mask_and_scale : boolean, optional
            Mask and scale during reading (default: False).
        """
        # dimension names
        self.dim = {"obs": obs_dim_name, "loc": loc_dim_name}

        # location names
        self.loc = {"ids": loc_ids_name, "descr": loc_descr_name}

        # time, time units and location
        self.var = {
            "time": time_var,
            "time_units": time_units,
            "lat": lat_var,
            "lon": lon_var,
            "alt": alt_var,
        }

        self.cache = cache
        self._cached = False
        self.mask_and_scale = mask_and_scale


class IRANcFile(RAFile):
    """
    Indexed ragged array file reader.
    """

    def __init__(self, filename, **kwargs):
        """
        Initialize.

        Parameters
        ----------
        filename : str
            Filename.
        loc_dim_name : str, optional
            Location dimension name (default: "locations").
        obs_dim_name : str, optional
            Observation dimension name (default: "time").
        loc_ids_name : str, optional
            Location IDs name (default: "location_id").
        loc_descr_name : str, optional
            Location description name (default: "location_description").
        time_units : str, optional
            Time units definition (default: "days since 1900-01-01 00:00:00").
        time_var : str, optional
            Time variable name (default: "time").
        lat_var : str, optional
            Latitude variable name (default: "lat").
        lon_var : str, optional
            Latitude variable name (default: "lon").
        alt_var : str, optional
            Altitude variable name (default: "alt").
        cache : boolean, optional
            Cache flag (default: False).
        mask_and_scale : boolean, optional
            Mask and scale during reading (default: False).
        """
        super().__init__(**kwargs)
        self.filename = filename

        # read location information
        with xr.open_dataset(self.filename,
                             mask_and_scale=self.mask_and_scale) as ncfile:
            var_list = [self.var["lon"], self.var["lat"], self.loc["ids"]]

            if self.cache:
                self.dataset = ncfile.load()
                self.locations = self.dataset[var_list].to_dataframe()
            else:
                self.dataset = None
                self.locations = ncfile[var_list].to_dataframe()

    @property
    def ids(self):
        """
        Location IDs property.

        Returns
        -------
        location_id : numpy.ndarray
            Location IDs.
        """
        return self.locations["location_id"]

    @property
    def lons(self):
        """
        Longitude coordinates property.

        Returns
        -------
        lon : numpy.ndarray
            Longitude coordinates.
        """
        return self.locations.lon

    @property
    def lats(self):
        """
        Latitude coordinates property.

        Returns
        -------
        lat : numpy.ndarray
            Latitude coordinates.
        """
        return self.locations.lat

    def read(self, location_id, variables=None):
        """
        Read a timeseries for a given location_id.

        Parameters
        ----------
        location_id : int
            Location_id to read.
        variables : list or None
            A list of parameter-names to read. If None, all parameters are read.
            If None, all parameters will be read. The default is None.

        Returns
        -------
        df : pandas.DataFrame
            A pandas.DataFrame containing the timeseries for the location_id.
        """
        pos = self.locations["location_id"] == location_id

        if not pos.any():
            print(f"location_id not found: {location_id}")
            data = None
        else:
            sel = self.locations[pos]
            i = sel.index.values[0]

            if self.cache:
                j = self.dataset["locationIndex"].values == i
                data = self.dataset.sel(locations=i, time=j)
            else:
                with xr.open_dataset(
                        self.filename,
                        mask_and_scale=self.mask_and_scale) as dataset:
                    j = dataset["locationIndex"].values == i
                    data = dataset.sel(locations=i, time=j)

            if variables is not None:
                data = data[variables]

        return data


class CRANcFile(RAFile):
    """
    Contiguous ragged array file reader.
    """

    def __init__(self, filename, row_var="row_size", **kwargs):
        """
        Initialize reader.

        Parameters
        ----------
        filename : str
            Filename.
        row_size : str, optional
            Row size variable name (default: "row_size")
        loc_dim_name : str, optional
            Location dimension name (default: "locations").
        obs_dim_name : str, optional
            Observation dimension name (default: "time").
        loc_ids_name : str, optional
            Location IDs name (default: "location_id").
        loc_descr_name : str, optional
            Location description name (default: "location_description").
        time_units : str, optional
            Time units definition (default: "days since 1900-01-01 00:00:00").
        time_var : str, optional
            Time variable name (default: "time").
        lat_var : str, optional
            Latitude variable name (default: "lat").
        lon_var : str, optional
            Latitude variable name (default: "lon").
        alt_var : str, optional
            Altitude variable name (default: "alt").
        cache : boolean, optional
            Cache flag (default: False).
        mask_and_scale : boolean, optional
            Mask and scale during reading (default: False).
        """
        super().__init__(**kwargs)
        self.var["row"] = row_var
        self.filename = filename

        # read location information
        with xr.open_dataset(self.filename,
                             mask_and_scale=self.mask_and_scale) as ncfile:
            var_list = [
                self.var["lon"],
                self.var["lat"],
                self.loc["ids"],
                self.var["row"],
            ]

            if self.cache:
                self.dataset = ncfile.load()

                self.locations = self.dataset[var_list].to_dataframe()
                self.locations[self.var["row"]] = np.cumsum(
                    self.locations[self.var["row"]])
            else:
                self.dataset = None

                self.locations = ncfile[var_list].to_dataframe()
                self.locations[self.var["row"]] = np.cumsum(
                    self.locations[self.var["row"]])

    @property
    def ids(self):
        """
        Location IDs property.

        Returns
        -------
        location_id : numpy.ndarray
            Location IDs.
        """
        return self.locations["location_id"]

    @property
    def lons(self):
        """
        Longitude coordinates property.

        Returns
        -------
        lon : numpy.ndarray
            Longitude coordinates.
        """
        return self.locations.lon

    @property
    def lats(self):
        """
        Latitude coordinates property.

        Returns
        -------
        lat : numpy.ndarray
            Latitude coordinates.
        """
        return self.locations.lat

    def read(self, location_id, variables=None):
        """
        Read a timeseries for a given location_id.

        Parameters
        ----------
        location_id : int
            Location_id to read.
        variables : list or None
            A list of parameter-names to read. If None, all parameters are read.
            If None, all parameters will be read. The default is None.

        Returns
        -------
        df : pandas.DataFrame
            A pandas.DataFrame containing the timeseries for the location_id.
        """
        pos = self.locations["location_id"] == location_id

        if not pos.any():
            print(f"location_id not found: {location_id}")
            data = None
        else:
            sel = self.locations[pos]
            i = sel.index.values[0]

            r_to = sel.row_size[i]
            if i > 0:
                r_from = int(self.locations.iloc[[i - 1]].row_size[i - 1])
            else:
                r_from = 0

            if self.cache:
                data = self.dataset.sel(locations=i, obs=slice(r_from, r_to))
            else:
                with xr.open_dataset(
                        self.filename,
                        mask_and_scale=self.mask_and_scale) as dataset:
                    data = dataset.sel(locations=i, obs=slice(r_from, r_to))

            if variables is not None:
                data = data[variables]

        return data

    def read_2d(self, variables=None):
        """
        (Draft!) Read all time series into 2d array.

        1d data: 1, 2, 3, 4, 5, 6, 7, 8
        row_size: 3, 2, 1, 2
        2d data:
        1 2 3 0 0
        4 5 0 0 0
        6 0 0 0 0
        7 8 0 0 0
        """
        row_size = np.array([3, 2, 1, 2])
        y = vrange(np.zeros_like(row_size), row_size)
        x = np.arange(row_size.size).repeat(row_size)
        target = np.zeros((4, 3))
        target[x, y] = np.arange(1, 9)
        print(target)


def vrange(starts, stops):
    """
    Create concatenated ranges of integers for multiple start/stop values.

    Parameters
    ----------
    starts : numpy.ndarray
        Starts for each range.
    stops : numpy.ndarray
        Stops for each range (same shape as starts).

    Returns
    -------
    ranges : numpy.ndarray
        Concatenated ranges.

    Example
    -------
        >>> starts = [1, 3, 4, 6]
        >>> stops  = [1, 5, 7, 6]
        >>> vrange(starts, stops)
        array([3, 4, 4, 5, 6])
    """
    stops = np.asarray(stops)
    l = stops - starts # Lengths of each range.
    return np.repeat(stops - l.cumsum(), l) + np.arange(l.sum())



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

from typing import Any, Union, Sequence

import numpy as np
import xarray as xr
from shapely.geometry.base import BaseGeometry

from pygeogrids import BasicGrid, CellGrid

from ascat.cf_array import cf_array_class, cf_array_type
from ascat.grids import GridRegistry
from ascat.utils import get_grid_gpis

from matplotlib import pyplot as plt


registry = GridRegistry()


@xr.register_dataset_accessor("pgg")
class PyGeoGriddedArrayAccessor:
    def __init__(self, xarray_obj: xr.Dataset):
        self._obj = xarray_obj
        self._grid = None

    @property
    def grid(self):
        if self._grid is None:
            if grid_name := self._obj.attrs.get("grid_mapping_name"):
                self._grid = registry.get(grid_name)
        assert isinstance(self._grid, BasicGrid)
        return self._grid

    def set_grid_name(
        self,
        grid_name: str,
        grid_class: Union[type, None] = None,
    ):
        try:
            self._grid = registry.get(grid_name)
        except KeyError:
            if grid_class is not None:
                registry.register(grid_name, grid_class)
            else:
                raise ValueError(
                    f"Grid {grid_name} is not registered."
                    " Please pass a class for creating a grid to the `grid` argument."
                )
        self._obj.attrs["grid_mapping_name"] = grid_name

    def sel_bbox(
            self,
            bbox: Sequence[float]
    ) -> xr.Dataset:
        """
        Select data for a bounding box.

        bbox : tuple, optional
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates.
        """
        gpis, lookup_vector = get_grid_gpis(
            grid=self.grid,
            bbox=bbox,
            return_lookup=True
        )
        return self._obj.pgg.sel_gpis(gpis, lookup_vector)

    def sel_coords(
        self,
        coords: Sequence[Sequence[float]],
        max_coord_dist: float = np.inf
    ) -> xr.Dataset:
        gpis, lookup_vector = get_grid_gpis(
            grid=self.grid,
            coords=coords,
            max_coord_dist=max_coord_dist,
            return_lookup=True,
        )
        return self._obj.pgg.sel_gpis(gpis, lookup_vector)

    def sel_cells(self, cells: Sequence[float]) -> xr.Dataset:
        assert isinstance(self.grid, CellGrid)
        gpis, lookup_vector = get_grid_gpis(
            grid=self.grid,
            cell=cells,
            return_lookup=True
        )
        return self._obj.pgg.sel_gpis(gpis, lookup_vector)

    def sel_geom(self, geom: BaseGeometry) -> xr.Dataset:
        gpis, lookup_vector = get_grid_gpis(
            grid=self.grid,
            geom=geom,
            return_lookup=True
        )
        return self._obj.pgg.sel_gpis(gpis, lookup_vector)

    def sel_gpis(
        self,
        gpis: Union[Sequence[int], None] = None,
        lookup_vector: Union[np.ndarray, None] = None,
    ) -> xr.Dataset:
        if lookup_vector is None:
            _, lookup_vector = get_grid_gpis(
                grid=self.grid,
                location_id=gpis,
                return_lookup=True
            )
        return self._obj.cf_geom.sel_instances(
            instance_vals=gpis,
            instance_lookup_vector=lookup_vector,
        )

    def lonlat_vars_from_gpi_var(
        self,
        gpi_var,
        lon_var="lon",
        lat_var="lat"
    ) -> tuple[xr.DataArray, xr.DataArray]:
        grid = self.grid
        lons, lats = grid.gpi2lonlat(self._obj[gpi_var])
        return (
            xr.DataArray(lons, dims=self._obj[gpi_var].dims, name=lon_var),
            xr.DataArray(lats, dims=self._obj[gpi_var].dims, name=lat_var)
        )



@xr.register_dataset_accessor("cf_geom")
class CFDiscreteGeometryAccessor:
    def __init__(self, xarray_obj: xr.Dataset):
        self._ds = xarray_obj
        self._obj = cf_array_class(self._ds, self.array_type)
        self._coord_vars = None
        self._instance_vars = None

    @property
    def array_type(self) -> str:
        return cf_array_type(self._ds)

    @property
    def timeseries_id(self) -> str:
        return self._obj.timeseries_id

    def set_coord_vars(self, coord_vars: Sequence[str]):
        self._coord_vars = coord_vars
        if self._instance_vars is None:
            self._instance_vars = []
        self._instance_vars.extend(coord_vars)

        self._obj = cf_array_class(
            self._ds,
            self.array_type,
            coord_vars=coord_vars,
            instance_vars=self._instance_vars,
        )

    def set_instance_vars(self, instance_vars: Sequence[str]):
        self._instance_vars = instance_vars
        self._obj = cf_array_class(
            self._ds,
            self.array_type,
            coord_vars=self._coord_vars,
            instance_vars=instance_vars,
        )

    def set_sample_dimension(self, sample_dim: str):
        self._ds = self._obj.set_sample_dimension(sample_dim)
        self._obj = cf_array_class(self._ds, self.array_type)
        return self._ds

    def sel_instances(
        self,
        instance_vals: Union[Sequence[int, str], np.ndarray, None] = None,
        instance_lookup_vector: Union[np.ndarray[Any, np.dtype[np.bool]], None] = None,
        **kwargs,
    ):
        return self._obj.sel_instances(
            instance_vals=instance_vals,
            instance_lookup_vector=instance_lookup_vector,
            **kwargs,
        )

    def to_point_array(self):
        return self._obj.to_point_array()

    def to_indexed_ragged(self, **kwargs):
        return self._obj.to_indexed_ragged(**kwargs)

    def to_contiguous_ragged(self, **kwargs):
        return self._obj.to_contiguous_ragged(**kwargs)

    def to_orthomulti(self, **kwargs):
        return self._obj.to_orthomulti(**kwargs)

    def resample_to_orthomulti(self, **kwargs):
        return self._obj.resample_to_orthomulti(**kwargs)

    def to_raster(self, *args, **kwargs):
        return self._obj.to_raster(*args, **kwargs)



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from datetime import timedelta
from functools import partial
from pathlib import Path

import dask
import numpy as np
import xarray as xr

from pyresample import kd_tree
from pyresample.geometry import AreaDefinition
from pyresample.geometry import SwathDefinition

from ascat.grids import GridRegistry

from ascat.utils import get_grid_gpis
from ascat.file_handling import Filenames
from ascat.file_handling import ChronFiles
import warnings

registry = GridRegistry()


class Swath(Filenames):
    """
    Class to read and merge swath files given one or more file paths.
    """

    def _read(self,
              filename,
              generic=True,
              preprocessor=None,
              **xarray_kwargs):
        """
        Open one swath file as an xarray.Dataset and preprocess it if necessary.

        Parameters
        ----------
        filename : str
            File to read.
        generic : bool, optional
            Not yet implemented, kept to match the signature.
        preprocessor : callable, optional
            Function to preprocess the dataset after opening.
        xarray_kwargs : dict
            Additional keyword arguments passed to xarray.open_dataset.

        Returns
        -------
        ds : xarray.Dataset
            Dataset.
        """
        ds = xr.open_dataset(
            filename,
            engine="h5netcdf",
            **xarray_kwargs,
        )
        if ds["location_id"].dtype != np.int32:
            ds["location_id"] = ds["location_id"].astype(np.int32)
        if preprocessor is not None:
            ds = preprocessor(ds)

        return ds

    def read(self, parallel=False, mask_and_scale=True, **kwargs):
        """
        Read the file or a subset of it.

        Parameters
        ----------
        parallel : bool, optional
            If True, read files in parallel.
        mask_and_scale : bool, optional
            If True, mask and scale the data.
        kwargs : dict
            Additional keyword arguments passed to `Filenames.read`.

        Returns
        -------
        ds : xarray.Dataset
            Dataset.
        """

        ds, closers = super().read(
            closer_attr="_close",
            parallel=parallel,
            mask_and_scale=mask_and_scale,
            **kwargs)
        if ds is not None:
            ds.set_close(partial(super()._multi_file_closer, closers))
            return ds

    @staticmethod
    def _nbytes(ds):
        return ds.nbytes

    def _merge(self, data):
        """
        Merge datasets.

        Parameters
        ----------
        data : list of xarray.Dataset
            Datasets to merge.

        Returns
        -------
        xarray.Dataset
            Merged dataset.
        """
        if data == []:
            return None

        merged_ds = xr.concat(
            [ds for ds in data if ds is not None and len(ds["obs"]) > 0],
            dim="obs",
            combine_attrs=self.combine_attributes,
            data_vars="minimal",
            coords="minimal",
        )

        return merged_ds

    @staticmethod
    def _ensure_obs(ds):
        """
        Makes sure that the sample dimension is named `obs`.
        """
        ds = ds.cf_geom.set_sample_dimension("obs")
        return ds

    @staticmethod
    def combine_attributes(attrs_list, context):
        """
        Decides which attributes to keep when merging swath files.

        Parameters
        ----------
        attrs_list : list of dict
            List of attributes dictionaries.
        context : None
            This currently is None, but will eventually be passed information about
            the context in which this was called.
            (see https://github.com/pydata/xarray/issues/6679#issuecomment-1150946521)

        Returns
        -------
        """
        # we don't need to pass on anything from global attributes, except for these
        global_attributes_to_pass_on_merge = [
            "grid_mapping_name", "featureType"
        ]
        if "global_attributes_flag" in attrs_list[0].keys():
            attrs_list[0].pop("global_attributes_flag")
            result = {}
            for attr in global_attributes_to_pass_on_merge:
                if val := attrs_list[0].get(attr, False):
                    result[attr] = val
            return result

        variable_attrs = attrs_list

        # this code taken straight from xarray/core/merge.py
        # Replicates the functionality of "drop_conflicts"
        # but just for variable attributes.
        result = {}
        dropped_keys = set()
        for attrs in variable_attrs:
            result.update({
                key: value
                for key, value in attrs.items()
                if key not in result and key not in dropped_keys
            })
            result = {
                key: value
                for key, value in result.items()
                if key not in attrs or
                xr.core.utils.equivalent(attrs[key], value)
            }
            dropped_keys |= {key for key in attrs if key not in result}
        return result


class SwathGridFiles(ChronFiles):
    """
    Class to manage chronological swath files with a date field in the filename.
    """

    def __init__(
        self,
        root_path,
        fn_templ,
        sf_templ,
        grid_name,
        date_field_fmt,
        cell_fn_format=None,
        cls_kwargs=None,
        err=True,
        fn_read_fmt=None,
        sf_read_fmt=None,
        fn_write_fmt=None,
        sf_write_fmt=None,
        preprocessor=None,
        postprocessor=None,
        cache_size=0,
    ):
        """
        Initialize SwathFiles class.

        Parameters
        ----------
        root_path : str
            Root path.
        fn_templ : str
            Filename template (e.g. "{date}_ascat.nc").
        sf_templ : dict, optional
            Subfolder template defined as dictionary (default: None).
        grid_name : str
            Name of the grid - must be registered in the grid registry.
        date_field_fmt : str
            Date field format (e.g. "%Y%m%d").
        cell_fn_format : str, optional
            String to use to format cell file names (e.g. "{:04d}.nc").
        cls_kwargs : dict, optional
            Class keyword arguments (default: None).
        err : bool, optional
            Set true if a file error should be re-raised instead of
            reporting a warning.
            Default: True
        fn_read_fmt : str or function, optional
            Filename format for read operation.
        sf_read_fmt : str or function, optional
            Subfolder format for read operation.
        fn_write_fmt : str or function, optional
            Filename format for write operation.
        sf_write_fmt : str or function, optional
            Subfolder format for write operation.
        preprocessor : callable, optional
            Function to preprocess datasets after opening.
        postprocessor : callable, optional
            Function to pass to the `postprocessor` argument of `ascat.cell.RaggedArrayTs.write`
            when stacking to cell files.
        cache_size : int, optional
            Number of files to keep in memory (default=0).
        """
        # first check if any files directly under root_path contain the ending (make
        # sure not to iterate through every file - just stop after the first one).
        # This allows the user to set the root path either at the place necessitated by
        # the sf_templ or directly at the level of the files. However, the user still
        # cannot set the root path anywhere else in the directory structure (e.g. within
        # a satellite but above a year). In order to choose a specific satellite, must
        # pass that as a fmt_kwarg
        ending = fn_templ.split(".")[-1]
        for f in Path(root_path).glob(f"*.{ending}"):
            if f.is_file():
                sf_templ = None
                sf_read_fmt = None
                break

        super().__init__(root_path, Swath, fn_templ, sf_templ, cls_kwargs, err,
                         fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt,
                         cache_size)

        self.date_field_fmt = date_field_fmt
        self.grid_name = grid_name
        self.grid = registry.get(grid_name)

        self.cell_fn_format = cell_fn_format
        self.preprocessor = preprocessor
        self.postprocessor = postprocessor

    @classmethod
    def from_product_id(
        cls,
        path,
        product_id,
    ):
        """Create a SwathGridFiles object based on a product_id.

        Returns a SwathGridFiles object initialized with an io_class specified
        by `product_id` (case-insensitive).

        Parameters
        ----------
        path : str or Path
            Path to the swath file collection.
        product_id : str
            Identifier for the specific ASCAT product the swath files are part of.

        Raises
        ------
        ValueError
            If product_id is not recognized.

        Examples
        --------
        >>> my_swath_collection = SwathFileCollection.from_product_id(
        ...     "/path/to/swath/files",
        ...     "H129",
        ... )

        """
        from ascat.product_info import swath_io_catalog
        product_id = product_id.upper()
        if product_id in swath_io_catalog:
            product_class = swath_io_catalog[product_id]
        else:
            error_str = f"Product {product_id} not recognized. Valid products are"
            error_str += f" {', '.join(swath_io_catalog.keys())}."
            raise ValueError(error_str)

        return cls.from_product_class(path, product_class)

    @classmethod
    def from_product_class(
        cls,
        path,
        product_class,
    ):
        """Create a SwathGridFiles from a given io_class.

        Returns a SwathGridFiles object initialized with the given io_class.

        Parameters
        ----------
        path : str or Path
            Path to the swath file collection.
        io_class : class
            Class to use for reading and writing the swath files.

        Examples
        --------
        >>> my_swath_collection = SwathFileCollection.from_io_class(
        ...     "/path/to/swath/files",
        ...     AscatH129Swath,
        ... )

        """
        return cls(
            path,
            product_class.fn_pattern,
            product_class.sf_pattern,
            grid_name=product_class.grid_name,
            cell_fn_format=product_class.cell_fn_format,
            date_field_fmt=product_class.date_field_fmt,
            fn_read_fmt=product_class.fn_read_fmt,
            sf_read_fmt=product_class.sf_read_fmt,
            preprocessor=product_class.preprocess_,
        )

    def _spatial_filter(
        self,
        filenames,
        cell=None,
        location_id=None,
        coords=None,
        bbox=None,
        geom=None,
    ):
        """
        Filter a search result for cells matching a spatial criterion.

        Parameters
        ----------
        cell : int or list of int
            Grid cell number to read.
        location_id : int or list of int
            Location id.
        coords : tuple of numeric or tuple of iterable of numeric
            Tuple of (lon, lat) coordinates.
        bbox : tuple
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates.

        Returns
        -------
        filenames : list of str
            Filenames.
        """

        if cell is not None:
            gpis = get_grid_gpis(self.grid, cell=cell)
            spatial = SwathDefinition(
                lats=self.grid.arrlat[gpis],
                lons=self.grid.arrlon[gpis],
            )
        elif location_id is not None:
            gpis = get_grid_gpis(self.grid, location_id=location_id)
            spatial = SwathDefinition(
                lats=self.grid.arrlat[gpis],
                lons=self.grid.arrlon[gpis],
            )
        elif coords is not None:
            spatial = SwathDefinition(
                lats=[coords[1]],
                lons=[coords[0]],
            )
        elif (bbox or geom) is not None:
            if bbox is not None:
                # AreaDefinition expects (lonmin, latmin, lonmax, latmax)
                # but bbox is (latmin, latmax, lonmin, lonmax)
                bbox = (bbox[2], bbox[0], bbox[3], bbox[1])
            else:
                # If we get a geometry just take its bounding box and check
                # that intersection.
                #
                # shapely.geometry.bounds is already in the correct order
                bbox = geom.bounds
            spatial = AreaDefinition(
                "bbox",
                "",
                "EPSG:4326",
                {
                    "proj": "latlong",
                    "datum": "WGS84"
                },
                1000,
                1000,
                bbox,
            )
        else:
            spatial = None

        if spatial is None:
            return filenames

        filtered_filenames = []
        for filename in filenames:
            lazy_result = dask.delayed(self._check_intersection)(filename,
                                                                 spatial)
            filtered_filenames.append(lazy_result)

        def none_filter(fname_list):
            return [l for l in fname_list if l is not None]

        filtered_filenames = dask.delayed(none_filter)(
            filtered_filenames).compute()

        return filtered_filenames

    def _check_intersection(self, filename, spatial):
        """
        Check if a file intersects with a pyresample SwathDefinition or AreaDefinition.

        Parameters
        ----------
        filename : str
            Filename.
        gpis : list of int
            List of gpis.

        Returns
        -------
        bool
            True if the file intersects with the gpis.
        """
        f = self.cls(filename)
        ds = f.read()
        with f.read() as ds:
            lons, lats = ds["longitude"].values, ds["latitude"].values
            swath_def = SwathDefinition(lats=lats, lons=lons)
            n_info = kd_tree.get_neighbour_info(
                swath_def,
                spatial,
                radius_of_influence=15000,
                neighbours=1,
            )
            valid_input_index, _, _ = n_info[:3]
        if np.any(valid_input_index):
            return filename
        return None

    def swath_search(
        self,
        dt_start,
        dt_end,
        dt_delta=None,
        search_date_fmt="%Y%m%d*",
        date_field="date",
        end_inclusive=True,
        cell=None,
        location_id=None,
        coords=None,
        bbox=None,
        geom=None,
        **fmt_kwargs,
    ):
        """
        Search for swath files within a time range and spatial criterion.

        Parameters
        ----------
        dt_start : datetime
            Start date.
        dt_end : datetime
            End date.
        dt_delta : timedelta
            Time delta.
        search_date_fmt : str
            Search date format.
        date_field : str
            Date field.
        end_inclusive : bool
            End date inclusive.
        cell : int or list of int
            Grid cell number to read.
        location_id : int or list of int
            Location id.
        coords : tuple of numeric or tuple of iterable of numeric
            Tuple of (lon, lat) coordinates.
        bbox : tuple
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates.
        geom : shapely.geometry
            Geometry.
        fmt_kwargs : dict
            Additional keyword arguments passed to ascat.file_handling.ChronFiles.search_period.

        Returns
        -------
        list of str
            Filenames.
        """
        dt_delta = dt_delta or timedelta(days=1)

        filenames = self.search_period(
            dt_start,
            dt_end,
            dt_delta,
            search_date_fmt,
            date_field,
            date_field_fmt=self.date_field_fmt,
            end_inclusive=end_inclusive,
            **fmt_kwargs,
        )

        filtered_filenames = self._spatial_filter(
            filenames,
            cell=cell,
            location_id=location_id,
            coords=coords,
            bbox=bbox,
            geom=geom,
        )

        if len(filtered_filenames) == 0:
            raise FileNotFoundError(
                f"No files found for {dt_start} to {dt_end} with the given spatial criteria."
            )

        return filtered_filenames

    def read(
        self,
        date_range,
        dt_delta=None,
        search_date_fmt="%Y%m%d*",
        date_field="date",
        end_inclusive=True,
        cell=None,
        location_id=None,
        coords=None,
        max_coord_dist=None,
        bbox=None,
        geom=None,
        read_kwargs=None,
        **fmt_kwargs,
    ):
        """
        Extract data from swath files within a time range and spatial criterion.

        Parameters
        ----------
        date_range : tuple of datetime.datetime
            Start and end date.
        dt_delta : timedelta
            Time delta.
        search_date_fmt : str
            Search date format.
        date_field : str
            Date field.
        end_inclusive : bool
            If True (default), include data from the end date in the result. Otherwise,
            exclude it.
        cell : int or list of int
            Grid cell number to read.
        location_id : int or list of int
            Location id to read.
        coords : tuple of numeric or tuple of iterable of numeric
            Tuple of (lon, lat) coordinates to read.
        max_coord_dist : float
            Maximum distance in meters to search for grid points near the given
            coordinates. If None, the default is np.inf.
        bbox : tuple
            Tuple of (latmin, latmax, lonmin, lonmax) coordinates to bound the data.
        geom : shapely.geometry
            Geometry to bound the data.

        Returns
        -------
        xarray.Dataset
            Dataset.
        """
        dt_start, dt_end = date_range
        filenames = self.swath_search(
            dt_start,
            dt_end,
            dt_delta,
            search_date_fmt,
            date_field,
            end_inclusive,
            **fmt_kwargs,
        )

        date_range = (np.datetime64(dt_start), np.datetime64(dt_end))

        read_kwargs = read_kwargs or {}

        def filter_ds_spatial(ds):
            if cell is not None:
                ds = ds.pgg.sel_cells(cell)
            elif location_id is not None:
                ds = ds.pgg.sel_gpis(location_id, gpi_var="location_id")
            elif coords is not None:
                ds = ds.pgg.sel_coords(coords, max_coord_dist=max_coord_dist)
            elif bbox is not None:
                ds = ds.pgg.sel_bbox(bbox)
            elif geom is not None:
                ds = ds.pgg.sel_geom(geom)
            return ds

        def preprocessor(ds):
            if self.preprocessor is not None:
                ds = self.preprocessor(ds)
            ds = filter_ds_spatial(ds)
            return ds

        read_kwargs["preprocessor"] = preprocessor

        data = self.cls(filenames).read(**read_kwargs)

        if data:
            if date_range is not None:
                mask = (data["time"] >= date_range[0]) & (
                    data["time"] <= date_range[1])
                data = data.sel(obs=mask.compute())

            return data
        warning_str = (
            "No data found for specified criteria, returning None:\n"
            f"date_range={date_range}\n"
            f"cell={cell}, location_id={location_id}, coords={coords}, bbox={bbox},\n"
            f"geom={geom}, max_coord_dist={max_coord_dist}, \n")
        warnings.warn(warning_str, UserWarning, 2)

    def stack_to_cell_files(
        self,
        out_dir,
        max_nbytes,
        date_range=None,
        fmt_kwargs=None,
        cells=None,
        print_progress=True,
        parallel=True,
    ):
        """
        Stack all swath files to cell files, writing them in parallel.

        Parameters
        ----------
        out_dir : str
            Output directory.
        max_nbytes : int
            Maximum number of bytes to open as xarray datasets before dumping to disk.
        date_range : tuple of datetime.datetime, optional
            Start and end date for the search.
        fmt_kwargs : dict, optional
            Additional keyword arguments passed to ascat.file_handling.ChronFiles.search_period.
        cells : list of int, optional
            List of grid cell numbers to read. If None (default), all cells are read.
        print_progress : bool, optional
            If True (default), print progress bars.
        parallel: bool, optional
            If True, write data to files in parallel (use all available resources).
        """
        from ascat.cell import RaggedArrayTs

        fmt_kwargs = fmt_kwargs or {}
        if date_range is not None:
            dt_start, dt_end = date_range
            filenames = self.swath_search(
                dt_start, dt_end, cell=cells, **fmt_kwargs)
        else:
            filenames = list(Path(self.root_path).glob("**/*.nc"))

        swath = self.cls(filenames)

        for ds in swath.iter_read_nbytes(
                max_nbytes,
                preprocessor=self.preprocessor,
                print_progress=print_progress,
                chunks=-1):
            ds_cells = self.grid.gpi2cell(ds["location_id"])
            if isinstance(ds_cells, np.ma.MaskedArray):
                ds_cells = ds_cells.compressed()
            ds_cells = xr.DataArray(ds_cells, dims="obs", name="cell")

            # sorting here enables us to manually select each cell's data much faster
            # than using a .groupby
            ds = ds.sortby(ds_cells)

            unique_cells, cell_counts = np.unique(ds_cells, return_counts=True)
            cell_counts = np.hstack([0, np.cumsum(cell_counts)])

            # for each cell in unique cells, isel the slice from the dataarray corresponding to it
            ds_list = []
            cell_fnames = []
            for i, c in enumerate(unique_cells):
                if (cells is None) or (c in cells):
                    cell_ds = ds.isel(
                        obs=slice(cell_counts[i], cell_counts[i + 1]))
                    if len(cell_ds) == 0:
                        continue
                    ds_list.append(cell_ds)
                    cell_fname = Path(out_dir) / self.cell_fn_format.format(c)
                    cell_fnames.append(cell_fname)

            writer_class = RaggedArrayTs(cell_fnames)
            writer_class.write(
                ds_list,
                parallel=parallel,
                postprocessor=self.postprocessor,
                ra_type="point",
                mode="a",
                print_progress=print_progress)

        if print_progress:
            print("\n")



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ascat.file_handling import Filenames
from ascat.utils import get_toi_subset, get_roi_subset
from ascat.utils import mask_dtype_nans


class AscatFile(Filenames):
    """
    Class reading ASCAT files.
    """
    def __init__(self, filename):
        """
        Initialize AscatFile.

        Parameters
        ----------
        filename : str
            Filename.
        """
        super().__init__(filename)

    def read(self, toi=None, roi=None, **kwargs):
        """
        Read ASCAT Level 1b data.

        Parameters
        ----------
        toi : tuple of datetime, optional
            Filter data for given time of interest (default: None).
            e.g. (datetime(2020, 1, 1, 12), datetime(2020, 1, 2))
        roi : tuple of 4 float, optional
            Filter data for region of interest (default: None).
            e.g. latmin, lonmin, latmax, lonmax

        Returns
        -------
        data : xarray.Dataset or numpy.ndarray
            ASCAT data.
        metadata : dict
            Metadata.
        """
        data, metadata = super().read(**kwargs)

        if toi:
            data = get_toi_subset(data, toi)

        if roi:
            data = get_roi_subset(data, roi)

        return data, metadata

    def read_period(self, dt_start, dt_end, **kwargs):
        """
        Read interval.

        Parameters
        ----------
        dt_start : datetime
            Start datetime.
        dt_end : datetime
            End datetime.

        Returns
        -------
        data : xarray.Dataset or numpy.ndarray
            ASCAT data.
        metadata : dict
            Metadata.
        """
        return self.read(toi=(dt_start, dt_end), **kwargs)



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Test download.
"""

import os
import unittest
from tempfile import mkdtemp
from datetime import datetime, timedelta

from ascat.download.connectors import HsafConnector
from ascat.download.connectors import EumConnector

credentials = {
    'EUM': {
        'consumer_key': os.getenv('EUM_CONSUMER_KEY'),
        'consumer_secret': os.getenv('EUM_CONSUMER_SECRET')
    },
    'HSAF': {
        'user': os.getenv('HSAF_FTP_USER'),
        'password': os.getenv('HSAF_FTP_PASSWORD')
    }
}


class TestDownload(unittest.TestCase):

    def setUp(self):
        """
        Get connection details.
        """
        self.local_path = mkdtemp()

        yesterday = (datetime.now() - timedelta(1))
        today = datetime.today()
        self.start_date = yesterday
        self.end_date = today

    @unittest.skipIf(credentials['EUM']['consumer_key'] is None,
                     "Skip EUMETSAT connection test")
    def test_eumetsat_connect(self):
        """
        Test EUMETSAT connection
        """
        connector = EumConnector()
        connector.connect(credentials['EUM'])

    @unittest.skipIf(credentials['EUM']['consumer_key'] is None,
                     "Skip EUMETSAT download test")
    def test_eumetsat_download(self):
        """
        Test EUMETSAT download.
        """
        product = "EO:EUM:DAT:METOP:SOMO12"

        connector = EumConnector()
        connector.connect(credentials['EUM'])
        connector.download(
            product, self.local_path, self.start_date, self.end_date, limit=1)

    @unittest.skipIf(credentials['HSAF']['user'] is None,
                     "Skip H SAF connection test")
    def test_hsaf_connect(self):
        """
        Test H SAF connection.
        """
        connector = HsafConnector()
        connector.connect(credentials['HSAF'])
        connector.close()

    @unittest.skipIf(credentials['HSAF']['user'] is None,
                     "Skip H SAF download test")
    def test_hsaf_download(self):
        """
        Test H SAF download.
        """
        remote_path = '/products/h08/h08_cur_mon_nc'

        connector = HsafConnector()
        connector.connect(credentials['HSAF'])
        connector.download(
            remote_path,
            self.local_path,
            self.start_date,
            self.end_date,
            limit=1)
        connector.close()


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
HTTP and FTP download module.
"""

import base64
import urllib
import requests
import logging
from pathlib import Path, WindowsPath, PosixPath
from ftplib import FTP
from datetime import timedelta
import concurrent.futures

from tqdm.auto import tqdm

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,
                    format="%(asctime)s %(levelname)s: %(message)s")


class Connector:
    """
    Base class for connecting and downloading from remote source.
    """

    def __init__(self, base_url):
        """
        Initialize connector.

        Parameters
        ----------
        base_url : string
            Location of remote resource.
        """
        self.base_url = base_url

    def connect(self, credentials):
        """
        Establish connection to remote source.

        Parameters
        ----------
        credentials : dict
            Dictionary of needed authentication parameters.
        """
        pass

    def download_date_range(self, path, start_date, end_date):
        """
        Fetch resource location for download of multiple files in date range.

        Parameters
        ----------
        path : string
            Local directory, where found datasets are stored.
        start_date : datetime
            Start date of date range interval.
        end_date : datetime
            End date of date range interval.
        """
        pass

    def download_file(self, file_remote, file_local):
        """
        Download single file from passed url to local file.

        Parameters
        ----------
        file_remote : string
            Path of file to download.
        file_local : string
            Path (local) where to save file.
        """
        pass

    def close(self):
        """
        Close connection.
        """
        pass


class HttpConnector(Connector):
    """
    Class for http requests.
    """

    def __init__(self, base_url):
        """
        Initialize connector.

        Parameters
        ----------
        base_url : string
            Location of remote resource.
        """
        self.base_url = base_url

    def download_file(self, file_remote, file_local, overwrite=False, n_retry=5):
        """
        Download single file from passed url to local file.

        Parameters
        ----------
        file_remote : string
            Path of file to download
        file_local : string
            Path (local) where to save file
        overwrite : bool, optional
            If True, existing files will be overwritten.
        """
        request_flag = False

        i = 0
        while i < n_retry:
            logging.debug("Send request")

            try:
                stream_response = requests.get(
                    file_remote,
                    params={"format": "json"},
                    stream=True,
                    headers={"Authorization": f"Bearer {self.access_token}"})

                self._assert_response(stream_response)
            except AssertionError:
                i += 1
                logging.debug(f"API Request failed - retry #{i}")
            else:
                logging.debug("API Request successful")
                request_flag = True
                break
        else:
            logging.debug("Maximum number of API requests failed. Abort.")

        if request_flag:
            total = int(stream_response.headers["content-length"])

            suffix = ""
            if stream_response.headers["Content-Type"] == "application/zip":
                suffix = ".zip"

            filename = file_local.parent / (file_local.name + suffix)

            if not overwrite and filename.exists():
                lstat = filename.lstat()
                if lstat.st_size == total:
                    logging.info("Skip download. File exits.")
            else:
                pbar = tqdm(desc=file_local.name,
                            total=total,
                            unit="B",
                            unit_divisor=1024,
                            unit_scale=True,
                            leave=False)

                with open(filename, "wb") as fp:
                    for chunk in stream_response.iter_content(chunk_size=1024):
                        if chunk:
                            fp.write(chunk)
                            fp.flush()
                            pbar.update(len(chunk))
                pbar.close()

                if filename.exists():
                    lstat = filename.lstat()
                    if lstat.st_size == total:
                        logging.debug("Download successful")
                    else:
                        logging.error("Download unsuccessful (file size mismatch)")
                else:
                    logging.error("Downloaded file not found")


class FtpConnector(Connector):
    """
    Class for downloading via FTP
    """

    def __init__(self, base_url):
        """
        Initialize connector.

        Parameters
        ----------
        base_url : string
            Location of remote resource.
        """
        super().__init__(base_url)
        self.ftp = FTP(self.base_url)

    def connect(self, credentials):
        """
        Establish connection to FTP source.

        Parameters
        ----------
        credentials : dict
            Dictionary of needed authentication parameters.
        """
        try:
            self.ftp.login(credentials["user"], credentials["password"])
            logging.info("FTP connection successfully established")
        except:
            logging.error("FTP connection failed. User or password incorrect")

    def download_file(self, file_remote, file_local, overwrite=False):
        """
        Download single file from passed url to local file

        Parameters
        ----------
        file_remote : string
            path of file to download
        file_local : string
            path (local) where to save file
        overwrite : bool, optional
            If True, existing files will be overwritten.
        """
        if file_remote not in self.ftp.nlst():
            logging.warning(f"File not accessible on FTP: {file_remote}")
        else:
            logging.debug(f"Start download: {file_remote}")

            total = self.ftp.size(file_remote)

            if not overwrite and file_local.exists():
                lstat = file_local.lstat()
                if lstat.st_size == total:
                    logging.info("Skip download. File exits.")
            else:
                pbar = tqdm(desc=file_local.name,
                            total=total,
                            unit="B",
                            unit_divisor=1024,
                            unit_scale=True,
                            dynamic_ncols=True,
                            leave=False)

                with open(file_local, "wb") as fp:

                    def cb(data):
                        pbar.update(len(data))
                        fp.write(data)

                    self.ftp.retrbinary(f"RETR {file_remote}", cb)

                pbar.close()

                if file_local.exists():
                    lstat = file_local.lstat()
                    if lstat.st_size == total:
                        logging.debug("Download successful")
                    else:
                        logging.error(
                            "Download unsuccessful (file size mismatch)")
                else:
                    logging.error("Downloaded file not found")

    def close(self):
        """
        Close connection.
        """
        self.ftp.close()
        logging.info("FTP disconnected")


class HsafConnector(FtpConnector):
    """
    Class for downloading from HSAF via FTP.
    """

    def __init__(self, base_url="ftphsaf.meteoam.it"):
        """
        Initialize connector.

        Parameters
        ----------
        base_url : string, optional
            Location of remote resource (default: ftphsaf.meteoam.it).
        """
        super().__init__(base_url)

    def download(self,
                 remote_path,
                 local_path,
                 start_date,
                 end_date,
                 limit=None,
                 overwrite=False):
        """
        Fetch resource location for download of multiple files in date range.

        Parameters
        ----------
        remote_path : string
            Remote directory, where found datasets are stored.
        local_path : string
            Local directory, where found datasets are stored.
        start_date : datetime
            Start date of date range interval.
        end_date : datetime
            End date of date range interval.
        limit : int, optional
            Filter used to limit the returned results (default: 1).
        overwrite : bool, optional
            If True, existing files will be overwritten.
        """
        download_url_list = []
        local_file_list = []

        local_path = str2path(local_path)

        i = 0
        for daily_files in self.files(remote_path, start_date, end_date):
            for file_remote in daily_files:
                file_local = local_path / file_remote
                local_file_list.append(file_local)
                download_url_list.append(file_remote)

                i = i + 1
                if limit and limit == i:
                    break
            else:
                # continue if the inner loop wasn't broken
                continue

            # inner loop was broken, break the outer
            break

        download_url_list, local_file_list = zip(*sorted(zip(
            download_url_list, local_file_list)))

        with tqdm(desc="Downloads", total=len(download_url_list)) as pbar:
            for download_url, local_file in zip(download_url_list,
                                                local_file_list):
                self.download_file(download_url,
                                   local_file,
                                   overwrite=overwrite)
                pbar.update()

    def files(self, remote_path, start_date, end_date):
        """
        Generator retrieving file list for given date range.

        Parameters
        ----------
        remote_path : string
            Remote directory, where found datasets are stored.
        local_path : string
            Local directory, where found datasets are stored.
        start_date : datetime
            Start date of date range interval.
        end_date : datetime
            End date of date range interval.

        Yields
        ------
        matches : list
            List of daily files.
        """
        self.ftp.cwd(remote_path)

        list_of_files = []
        self.ftp.retrlines("NLST ", list_of_files.append)

        days = end_date - start_date
        for i in range(days.days):
            date = (start_date + timedelta(days=i)).strftime("%Y%m%d")
            matches = sorted([x for x in list_of_files if date in x],
                             reverse=True)
            yield matches


class EumConnector(HttpConnector):
    """
    Class for downloading from EUMETSAT via HTTP requests.
    """

    def __init__(self, base_url="https://api.eumetsat.int"):
        """
        Initialize connector.

        Parameters
        ----------
        base_url : string, optional
            Location of remote resource (default: https://api.eumetsat.int).
        """
        super().__init__(base_url)

    def connect(self, credentials):
        """
        Establish connection to EUMETSAT.

        Parameters
        ----------
        credentials: dict
            Dictionary of needed authentication parameters.
        """
        self.access_token = self._generate_token(
            consumer_key=credentials["consumer_key"],
            consumer_secret=credentials["consumer_secret"])

    def download(self,
                 product,
                 local_path,
                 start_date,
                 end_date,
                 max_workers=1,
                 coords=None,
                 limit=None):
        """
        Fetch resource location for download of multiple files in daterange.

        Parameters
        ----------
        product : string
            Product.
        coords : list
            Coordinates of polygon, where files will be downloaded
        local_path : string
            Local directory, where found datasets are stored.
        start_date : datetime
            Start date of date range interval.
        end_date : datetime
            End date of date range interval.
        max_workers : int, optional
            Number of parallel downloads (default: 1).
        coords : list of float, optional
            A custom polygon using EPSG:4326 decimal degrees (default: None).
        limit : int, optional
            Filter used to limit the returned results (default: None).
        """
        local_path = str2path(local_path)

        service_search = f"{self.base_url}/data/search-products/os"
        service_download = f"{self.base_url}/data/download/"

        dataset_parameters = {"format": "json", "pi": product}

        fmt = "%Y-%m-%dT%H:%M:%S.%fZ"
        dataset_parameters["dtstart"] = start_date.strftime(fmt)
        dataset_parameters["dtend"] = end_date.strftime(fmt)

        if coords:
            dataset_parameters["geo"] = "POLYGON(({}))".format(",".join(
                [f"{coord[0]} {coord[1]}" for coord in coords]))

        url = service_search
        response = requests.get(url, dataset_parameters)
        found_data_sets = response.json()

        url = service_search
        dataset_parameters['si'] = 0
        items_per_page = 10

        if "type" in found_data_sets:
            if found_data_sets["type"] == "ExceptionReport":
                msg = found_data_sets["exceptions"][0]["exceptionText"]
                raise RuntimeError(msg)

        all_found_data_sets = []
        while dataset_parameters['si'] < found_data_sets['totalResults']:
            response = requests.get(url, dataset_parameters)
            found_data_sets = response.json()
            all_found_data_sets.append(found_data_sets)
            dataset_parameters[
                'si'] = dataset_parameters['si'] + items_per_page

        download_url_list = []
        local_file_list = []

        if all_found_data_sets:

            for found_data_sets in all_found_data_sets:
                for selected_data_set in found_data_sets["features"]:
                    coll_id = selected_data_set["properties"][
                        "parentIdentifier"]
                    product_id = selected_data_set["properties"]["identifier"]
                    url_temp = f"collections/{coll_id}/products/{product_id}"
                    download_url_list.append(service_download +
                                             urllib.parse.quote(url_temp))
                    local_file_list.append(local_path / product_id)

            print(f"Found {found_data_sets['totalResults']} data sets")

            if limit and len(download_url_list) > limit:
                print(f"Limited to {limit} data sets")
                download_url_list = download_url_list[:limit]
                local_file_list = local_file_list[:limit]

            download_url_list, local_file_list = zip(*sorted(zip(
                download_url_list, local_file_list)))

            concurrent_download(self.download_file, download_url_list,
                                local_file_list, max_workers)
        else:
            print("No data sets found")

    def _generate_token(self, consumer_key, consumer_secret):
        """
        Function to generate an access token for interacting with
        EUMETSAT Data Service APIs.

        Parameters
        ----------
        consumer_key : str
            The consumer key as a string
        consumer_secret : str
            The consumer secret as a string.

        Returns
        -------
        access_token : str
            An access token (if pass) or None (if fail).
        """
        token_url = f"{self.base_url}/token"
        userpass = f"{consumer_key}:{consumer_secret}"
        encoded_userpass = base64.b64encode(userpass.encode()).decode()
        headers = {"Authorization": f"Basic {encoded_userpass}"}
        data_payload = {"grant_type": "client_credentials"}
        response = requests.post(token_url, headers=headers, data=data_payload)

        self._assert_response(response)

        return response.json()["access_token"]

    def _assert_response(self, response, success_code=200):
        """
        Function to check API key generation response. Will return an error
        if the key retrieval was not successful.

        Parameters
        ----------
        response : obj
            The authentication response.
        success_code : int, optional
            The expected sucess code (default: 200).

        Returns
        -------
        result : None or str
            Nothing if success, error message if fail.
        """
        msg = f"API Request Failed: {response.status_code}\n{response.content}"
        assert (response.status_code == success_code), msg


def concurrent_download(download_func,
                        download_url_list,
                        local_file_list,
                        max_workers=1):
    """
    Threaded file download.

    Parameters
    ----------
    download_func : function
        Download function.
    download_url_list : list
        Download URLs.
    local_file_list : list
        Local filenames.
    max_workers : int, optional
        Number of concurrent downloads (default: 1).
    """
    if max_workers == 1:
        with tqdm(desc="Downloads", total=len(download_url_list)) as pbar:
            for download_url, local_file in zip(download_url_list,
                                                local_file_list):
                download_func(download_url, local_file)
                pbar.update()
    else:
        with tqdm(desc="Downloads", total=len(download_url_list)) as pbar:
            with concurrent.futures.ThreadPoolExecutor(
                    max_workers=max_workers) as executor:

                future_down = {
                    executor.submit(download_func, url, local_file): url
                    for url, local_file in zip(download_url_list,
                                               local_file_list)
                }

                for future in concurrent.futures.as_completed(future_down):
                    _ = future_down[future]
                    pbar.update()


def str2path(path):
    """
    Convert str path to pathlib.Path object.

    Parameters
    ----------
    path : str, pathlib.Path
        Path.

    Returns
    -------
    path : pathlib.Path
        Pathlib path.
    """
    if not isinstance(path, (WindowsPath, PosixPath)):
        path = Path(path)

    return path
